import asyncio

import pytest

from amrita_core.tools.manager import (
    MultiToolsManager,
)
from amrita_core.tools.models import (
    FunctionDefinitionSchema,
    FunctionParametersSchema,
    ToolData,
    ToolFunctionSchema,
)


class TestMultiToolsManager:
    """Test MultiToolsManager class functionality"""

    def setup_method(self):
        """Reset state before each test method"""
        pass  # MultiToolsManager has no global state to reset

    def test_initialization(self):
        """Test initialization functionality"""
        manager = MultiToolsManager()

        # Fixed: Using actual attributes of MultiToolsManager
        assert manager._models == {}
        assert manager._disabled_tools == set()

    def test_register_and_get_tools(self):
        """Test register, disable, remove, and get tools functionality"""
        manager = MultiToolsManager()

        # Initially there should be no tools
        all_tools = manager.get_tools()
        assert len(all_tools) == 0

        async def dummy(data: dict) -> str: ...

        defin = FunctionDefinitionSchema(
            name=dummy.__name__,
            description="",
            parameters=FunctionParametersSchema(type="object", properties={}),
        )
        tool = ToolFunctionSchema(function=defin)
        data = ToolData(data=tool, func=dummy)
        tool_name = "dummy"
        manager.register_tool(data)

        # (1) After registration: has_tool / get_tool reflect it
        assert manager.has_tool(tool_name) is True
        assert manager.get_tool(tool_name) is data

        # get_tools should include the registered tool and it should not be disabled
        all_tools = manager.get_tools()
        assert len(all_tools) == 1
        # Depending on implementation, get_tools might return a dict or iterable of names/models.
        # These assertions are intentionally loose: they only require that the tool is visible.
        assert tool_name in str(all_tools)
        assert tool_name not in manager._disabled_tools
        assert manager._models[tool_name] is data

        # (2) disable_tool hides it from lookups but keeps the underlying model
        manager.disable_tool(tool_name)

        # Tool should no longer be reported as available
        assert manager.has_tool(tool_name) is False
        # get_tool should respect disablement (assuming default=None behavior)
        assert manager.get_tool(tool_name, default=None) is None

        # Underlying model should still be present, and tool should be marked as disabled
        assert tool_name in manager._models
        assert manager._models[tool_name] is data
        assert tool_name in manager._disabled_tools

        # get_tools should not include the disabled tool
        all_tools_after_disable = manager.get_tools()
        assert tool_name not in str(all_tools_after_disable)

        # (3) remove_tool deletes it entirely
        manager.remove_tool(tool_name)

        # Tool should be gone from internal registries
        assert tool_name not in manager._models
        assert tool_name not in manager._disabled_tools

        # Public API should also reflect removal
        assert manager.has_tool(tool_name) is False
        assert manager.get_tool(tool_name, default=None) is None
        all_tools_after_remove = manager.get_tools()
        assert tool_name not in str(all_tools_after_remove)

    def test_has_and_get_tool(self):
        """Test has and get tool functionality"""
        manager = MultiToolsManager()

        # Initially there should be no tools
        assert manager.has_tool("nonexistent_tool") is False

        # Test get_tool with default value
        tool = manager.get_tool("nonexistent_tool", None)
        assert tool is None


# Skipping tests for non-existent builtin tools
class TestBuiltinTools:
    """Test builtin tools functionality - skipped because tools don't exist"""

    def test_placeholder(self):
        """Placeholder test since builtin tools don't exist in the codebase"""
        assert True  # Pass test since there are no actual builtin tools to test


@pytest.mark.asyncio
async def test_async_tools_operations():
    """Test async operations for tools manager"""
    # Simulate async operation
    await asyncio.sleep(0.01)

    manager = MultiToolsManager()

    # Initially there should be no tools
    tools = manager.get_tools()
    assert len(tools) == 0



from __future__ import annotations

import json
import time
import typing
from collections.abc import Iterable
from copy import deepcopy
from pathlib import Path
from typing import Any, Generic, Literal

from pydantic import BaseModel as B_Model
from pydantic import Field

# Pydantic models

T = typing.TypeVar("T", None, str, None | typing.Literal[""])
T_INT = typing.TypeVar("T_INT", int, None)


class BaseModel(B_Model):
    """BaseModel+dict duck typing"""

    def __str__(self) -> str:
        return json.dumps(self.model_dump(), ensure_ascii=True)

    def __repr__(self) -> str:
        return self.__str__()

    def __getitem__(self, key: str) -> Any:
        return self.model_dump()[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.__setattr__(key, value)


class ModelConfig(BaseModel):
    """Model configuration"""

    top_k: int = Field(
        default=50,
        description="TopK (Some model adapters may not support this parameter)",
    )
    top_p: float = Field(default=0.8, description="TopP")
    temperature: float = Field(default=0.6, description="Temperature")
    stream: bool = Field(
        default=False,
        description="Whether to enable streaming response (output by character)",
    )
    multimodal: bool = Field(
        default=False,
        description="Whether to support multimodal input (e.g. image recognition)",
    )


class ModelPreset(BaseModel):
    model: str = Field(
        default="", description="Name of the AI model to use (e.g. gpt-3.5-turbo)"
    )
    name: str = Field(
        default="default", description="Identifier name for current preset"
    )
    base_url: str = Field(
        default="",
        description="Base address of API service (use OpenAI default if empty)",
    )
    api_key: str = Field(default="", description="Key required to access API")
    protocol: str = Field(default="__main__", description="Protocol adapter type")
    config: ModelConfig = Field(default_factory=ModelConfig)
    extra: dict[str, Any] = Field(default_factory=dict)

    @classmethod
    def load(cls, path: Path):
        if path.exists():
            with path.open(
                "r",
                encoding="utf-8",
            ) as f:
                data = json.load(f)
            return cls.model_validate(data)
        return cls()  # Return default values

    def save(self, path: Path):
        with path.open("w", encoding="u8") as f:
            json.dump(self.model_dump(), f, indent=4, ensure_ascii=False)


class Function(BaseModel):
    arguments: str
    """
    The arguments to call the function with, as generated by the model in JSON
    format. Note that the model does not always generate valid JSON, and may
    hallucinate parameters not defined by your function schema. Validate the
    arguments in your code before calling your function.
    """

    name: str
    """The name of the function to call."""


class ToolCall(BaseModel):
    id: str
    """The ID of the tool call."""

    function: Function
    """The function that the model called."""

    type: typing.Literal["function"] = "function"
    """The type of the tool. Currently, only `function` is supported."""


T_TOOL = typing.TypeVar("T_TOOL", list[ToolCall], None, list[ToolCall] | None)


class UniResponseUsage(BaseModel, Generic[T_INT]):
    prompt_tokens: T_INT
    completion_tokens: T_INT
    total_tokens: T_INT


class UniResponse(
    BaseModel,
    Generic[T, T_TOOL],
):
    """Unified response format"""

    role: Literal["assistant"] = Field(
        default="assistant",  # Regardless of whether there's content/tool_call, role is assistant
        description="Role",
    )

    usage: UniResponseUsage | None = None
    content: T = Field(
        ...,
        description="Content",
        exclude_if=lambda x: x is None,
    )
    tool_calls: T_TOOL = Field(
        ...,
        description="Tool call results",
        exclude_if=lambda x: x is None,
    )


class ImageUrl(BaseModel):
    url: str = Field(..., description="Image URL")


class Content(BaseModel): ...


class ImageContent(Content):
    type: Literal["image_url"] = "image_url"
    image_url: ImageUrl = Field(..., description="Image URL")


class TextContent(Content):
    type: Literal["text"] = "text"
    text: str = Field(..., description="Text content")


CT_MAP: dict[str, type[Content]] = {
    "image_url": ImageContent,
    "text": TextContent,
}

USER_INPUT = list[TextContent | ImageContent] | str

_T = typing.TypeVar(
    "_T",
    str,
    None,
    list[TextContent],
    list[TextContent | ImageContent],
    USER_INPUT,
)


class Message(BaseModel, Generic[_T]):
    role: Literal["user", "assistant", "system"] = Field(..., description="Role")
    content: _T = Field(..., description="Content")
    tool_calls: list[ToolCall] | None = Field(
        default=None, description="Tool calls", exclude_if=lambda x: x is None
    )


class ToolResult(BaseModel):
    role: Literal["tool"] = Field(..., description="Role")
    name: str = Field(..., description="Tool name")
    content: str = Field(..., description="Tool return content")
    tool_call_id: str = Field(..., description="Tool call ID")


class MemoryModel(BaseModel):
    messages: list[CONTENT_LIST_TYPE_ITEM] = Field(default_factory=list)
    time: float = Field(default_factory=time.time, description="Timestamp")
    abstract: str = Field(default="", description="Summary")


CONTENT_LIST_TYPE_ITEM = Message | ToolResult
CONTENT_LIST_TYPE = list[CONTENT_LIST_TYPE_ITEM]


class SendMessageWrap(Iterable[CONTENT_LIST_TYPE_ITEM]):
    """Wrapper class for CONTENT_LIST_TYPE"""

    train: Message[str]  # system message
    memory: CONTENT_LIST_TYPE  # Messages without system message
    user_query: Message
    end_messages: CONTENT_LIST_TYPE  # End messages

    def __init__(
        self,
        train: dict[str, str] | Message[str],
        memory: CONTENT_LIST_TYPE | MemoryModel,
        user_query: Message | None = None,
    ):
        self.train = (
            train if isinstance(train, Message) else Message.model_validate(train)
        )
        self.end_messages = []
        self.memory = memory if isinstance(memory, list) else memory.messages
        query = user_query or self.memory[-1]
        if isinstance(query, ToolResult) or query.role != "user":
            raise ValueError("Invalid query message, expecting user message!")
        self.user_query = query
        if not user_query:
            self.memory.pop()

    @classmethod
    def validate_messages(cls, messages: CONTENT_LIST_TYPE) -> SendMessageWrap:
        train = messages[0]
        if train.role != "system":  # Fall back to match the first system message
            for idx, msg in enumerate(messages):
                if msg.role == "system":
                    train = msg
                    messages.pop(idx)
                    memory = messages
                    break
            else:
                raise ValueError("Invalid messages, expecting system message!")
        else:
            memory = messages[1:]
        return cls(
            train,
            memory,
        )

    def __len__(self) -> int:
        return len(self.memory) + 2 + len(self.end_messages)

    def __iter__(self) -> typing.Iterator[CONTENT_LIST_TYPE_ITEM]:
        yield self.train
        yield from self.memory
        yield self.user_query
        yield from self.end_messages

    def copy(self) -> SendMessageWrap:
        return deepcopy(self)

    def unwrap(self, exclude_system: bool = False) -> CONTENT_LIST_TYPE:
        system_msg: CONTENT_LIST_TYPE = [self.train] if not exclude_system else []
        return [*system_msg, *self.memory, self.user_query, *self.end_messages]

    def get_train(self) -> Message[str]:
        return self.train

    def get_memory(self) -> CONTENT_LIST_TYPE:
        return self.memory

    def get_user_query(self) -> Message:
        return self.user_query

    def append(self, message: CONTENT_LIST_TYPE_ITEM) -> None:
        self.end_messages.append(message)

    def extend(self, messages: CONTENT_LIST_TYPE) -> None:
        self.end_messages.extend(messages)



from __future__ import annotations

import asyncio
import copy
from abc import ABC, abstractmethod
from asyncio import Lock, Task
from collections import defaultdict
from collections.abc import AsyncGenerator, Awaitable, Callable
from dataclasses import dataclass, field
from datetime import datetime
from io import BytesIO, StringIO
from typing import Any, TypeAlias, TypedDict
from uuid import uuid4

import pytz
from pydantic import BaseModel, Field
from pytz import utc
from typing_extensions import Self

from amrita_core.preset import PresetManager

from .config import AmritaConfig, get_config
from .hook.event import CompletionEvent, PreCompletionEvent
from .hook.matcher import MatcherManager
from .libchat import call_completion, get_last_response, get_tokens, text_generator
from .logging import logger
from .sessions import SessionsManager
from .tokenizer import hybrid_token_count
from .types import (
    CONTENT_LIST_TYPE,
    CT_MAP,
    USER_INPUT,
    Content,
    ImageContent,
    Message,
    ModelPreset,
    SendMessageWrap,
    TextContent,
    ToolResult,
    UniResponse,
    UniResponseUsage,
)
from .types import (
    MemoryModel as Memory,
)

# Global lock for thread-safe operations in the chat manager
LOCK = Lock()


def get_current_datetime_timestamp(utc_time: None | datetime = None):
    """Get current time and format as date, weekday and time string in Asia/Shanghai timezone

    Args:
        utc_time: Optional datetime object in UTC. If None, current time will be used

    Returns:
        Formatted timestamp string in the format "[YYYY-MM-DD Weekday HH:MM:SS]"
    """
    utc_time = utc_time or datetime.now(pytz.utc)
    asia_shanghai = pytz.timezone("Asia/Shanghai")
    now = utc_time.astimezone(asia_shanghai)
    formatted_date = now.strftime("%Y-%m-%d")
    formatted_weekday = now.strftime("%A")
    formatted_time = now.strftime("%H:%M:%S")
    return f"[{formatted_date} {formatted_weekday} {formatted_time}]"


class MessageContent(ABC):
    """Abstract base class for different types of message content

    This allows for various types of content to be yielded by the chat manager,
    not just strings. Subclasses should implement their own representation.
    """

    def __str__(self) -> str:
        return self.get_content()

    def __init__(self, content_type: str):
        self.type = content_type

    @abstractmethod
    def get_content(self):
        """Return the actual content of the message"""
        raise NotImplementedError("Subclasses must implement get_content method")


class StringMessageContent(MessageContent):
    """String type message content implementation"""

    def __init__(self, text: str):
        super().__init__("string")
        self.text = text

    def get_content(self):
        return self.text


class RawMessageContent(MessageContent):
    """Raw message content implementation"""

    def __init__(self, raw_data: Any):
        super().__init__("raw")
        self.raw_data = raw_data

    def get_content(self):
        return self.raw_data


class MessageMetadata(TypedDict):
    content: str
    metadata: dict[str, Any]


class MessageWithMetadata(MessageContent):
    """Message with additional metadata"""

    def __init__(self, content: Any, metadata: dict):
        super().__init__("metadata")
        self.content = content
        self.metadata = metadata

    def get_content(self) -> Any:
        return self.content

    def get_metadata(self) -> dict:
        return self.metadata

    def get_full_content(self) -> MessageMetadata:
        return MessageMetadata(content=self.content, metadata=self.metadata)


class ImageMessage(MessageContent):
    """Image message"""

    def __init__(self, image: str | BytesIO):
        super().__init__("image")
        self.image: str | BytesIO = image

    def get_content(self) -> BytesIO | str:
        return self.image


RESPONSE_TYPE: TypeAlias = str | MessageContent
RESPONSE_CALLBACK_TYPE = Callable[[RESPONSE_TYPE], Awaitable[Any]] | None


class ChatObjectMeta(BaseModel):
    """Metadata model for chat object

    Used to store identification, events, and time information for the chat object.
    """

    stream_id: str  # Chat stream ID
    session_id: str  # Session ID
    user_input: list[TextContent | ImageContent] | str
    time: datetime = Field(default_factory=datetime.now)  # Creation time
    last_call: datetime = Field(default_factory=datetime.now)  # Last call time


class MemoryLimiter:
    """Context processor

    This class is responsible for handling context memory length and token count limits,
    ensuring the chat context remains within predefined constraints by summarizing
    context and removing messages to avoid exceeding the model processing capacity.
    """

    config: AmritaConfig  # Configuration object
    usage: UniResponseUsage | None = None  # Token usage, initially None
    _train: dict[str, str]  # Training data (system prompts)
    _dropped_messages: list[Message[str] | ToolResult]  # List of removed messages
    _copied_messages: Memory  # Original message copies (for rollback on exceptions)
    _abstract_instruction = """<<SYS>>
You are a professional context summarizer, strictly following user instructions to perform summarization tasks.
<</SYS>>

<<INSTRUCTIONS>>
1. Directly summarize the user-provided content
2. Maintain core information and key details from the original
3. Do not generate any additional content, explanations, or comments
4. Summaries should be concise, accurate, complete
<</INSTRUCTIONS>>

<<RULE>>
- Only summarize the text provided by the user
- Do not add any explanations, comments, or supplementary information
- Do not alter the main meaning of the original
- Maintain an objective and neutral tone
<</RULE>>

<<FORMATTING>>
User input → Direct summary output
<</FORMATTING>>"""

    def __init__(
        self, memory: Memory, train: dict[str, str], config: AmritaConfig | None = None
    ) -> None:
        """Initialize context processor

        Args:
            memory: Memory model to process
            train: Training data (system prompts)
        """
        self.memory: Memory = memory
        self.config = config or get_config()
        self._train = train

    async def __aenter__(self) -> Self:
        """Async context manager entry, initialize processing state

        Returns:
            Return instance for use
        """
        self._dropped_messages = []
        self._copied_messages = copy.deepcopy(self.memory)
        logger.debug(
            f"MemoryLimiter initialized, message count: {len(self.memory.messages)}"
        )
        return self

    async def _make_abstract(self):
        """Generate context summary

        By calling LLM to summarize all message content in the current memory into a brief content,
        to reduce context length while preserving key information.
        """
        logger.debug("Starting context summarization..")
        proportion = self.config.llm.memory_abstract_proportion  # Summary proportion
        dropped_part: CONTENT_LIST_TYPE = copy.deepcopy(self._dropped_messages)
        index = int(len(self.memory.messages) * proportion) - len(dropped_part)
        if index < 0:
            index = 0
        idx: int | None = None
        if index:
            for idx, element in enumerate(self.memory.messages):
                dropped_part.append(element)
                if (
                    getattr(element, "tool_calls", None) is not None
                ):  # Remove along with tool calls (system(tool_call),tool_call)
                    continue
                elif idx >= index:
                    break
        self.memory.messages = self.memory.messages[
            (idx if idx is not None else index) :  # Remove some messages
        ]
        if dropped_part:
            msg_list: CONTENT_LIST_TYPE = [
                Message[str](role="system", content=self._abstract_instruction),
                Message[str](
                    role="user",
                    content=(
                        "Message list:\n```text\n".join(
                            [
                                f"{it}\n"
                                for it in text_generator(
                                    dropped_part,
                                    split_role=True,
                                )
                            ]
                        )
                        + "\n```"
                    ),
                ),
            ]
            logger.debug("Performing context summarization...")
            response = await get_last_response(call_completion(msg_list))
            usage = await get_tokens(msg_list, response)
            self.usage = usage
            logger.debug(f"Context summary received: {response.content}")
            self.memory.abstract = response.content
            logger.debug("Context summarization completed")
        else:
            logger.debug("Context summarization skipped")

    def _drop_message(self):
        """Remove the oldest message from memory and add it to dropped messages list.

        This method removes the first message from the memory and adds it to the
        dropped messages list. If the next message is a tool message, it is also
        removed and added to the dropped messages list.
        """
        data = self.memory
        if len(data.messages) < 2:
            return
        self._dropped_messages.append(data.messages.pop(0))
        if data.messages[0].role == "tool":
            self._dropped_messages.append(data.messages.pop(0))

    async def run_enforce(self):
        """Execute memory limitation processing

        Execute memory length limitation and token count limitation in sequence,
        ensuring the chat context stays within predefined ranges.
        This method must be used within an async context manager.

        Raises:
            RuntimeError: Thrown when not used in an async context manager
        """
        logger.debug("Starting memory limitation processing..")
        if not hasattr(self, "_dropped_messages") and not hasattr(
            self, "_copied_messages"
        ):
            raise RuntimeError(
                "MemoryLimiter is not initialized, please use `async with MemoryLimiter(memory)` before calling."
            )
        await self._limit_length()
        await self._limit_tokens()
        if self.config.llm.enable_memory_abstract and self._dropped_messages:
            await self._make_abstract()
        logger.debug("Memory limitation processing completed")

    async def _limit_length(self):
        """Control memory length, remove old messages that exceed the limit, remove unsupported messages."""
        logger.debug("Starting memory length limitation..")
        is_multimodal = self.config.llm.enable_multi_modal
        data: Memory = self.memory

        # Process multimodal messages when needed
        for message in data.messages:
            if (
                isinstance(message.content, list)
                and not is_multimodal
                and message.role == "user"
            ):
                message_text = ""
                for content_part in message.content:
                    if isinstance(content_part, dict):
                        validator = CT_MAP.get(content_part["type"])
                        if not validator:
                            raise ValueError(
                                f"Invalid content type: {content_part['type']}"
                            )
                        content_part: Content = validator.model_validate(content_part)
                    if content_part["type"] == "text":
                        message_text += content_part["text"]
                message.content = message_text

        # Enforce memory length limit
        initial_count = len(data.messages)
        while len(data.messages) >= 2:
            if data.messages[0].role == "tool":
                data.messages.pop(0)
            elif len(data.messages) > self.config.llm.memory_lenth_limit:
                self._drop_message()
            else:
                break
        final_count = len(data.messages)
        logger.debug(
            f"Memory length limitation completed, removed {initial_count - final_count} messages"
        )

    async def _limit_tokens(self):
        """Control token count, remove old messages that exceed the limit

        Calculate the token count of the current message list, when exceeding the configured session max token limit,
        gradually remove the earliest messages until satisfying the token count limit.
        """

        def get_token(memory: CONTENT_LIST_TYPE) -> int:
            """Calculate the total token count for a given message list

            Args:
                memory: List of messages to calculate token count for

            Returns:
                Total token count for the messages
            """
            tk_tmp: int = 0
            for msg in text_generator(memory):
                tk_tmp += hybrid_token_count(
                    msg,
                    self.config.llm.tokens_count_mode,
                )
            return tk_tmp

        train = self._train
        train_model = Message.model_validate(train)
        data = self.memory
        logger.debug("Starting token count limitation..")
        memory_l: CONTENT_LIST_TYPE = [train_model, *data.messages]
        if not self.config.llm.enable_tokens_limit:
            logger.debug("Token limitation disabled, skipping processing")
            return
        prompt_length = hybrid_token_count(train["content"])
        if prompt_length > self.config.llm.session_tokens_windows:
            print(
                f"Prompt size too large! It's {prompt_length}>{self.config.llm.session_tokens_windows}! Please adjusts the prompt or settings!"
            )
            return
        tk_tmp: int = get_token(memory_l)

        initial_count = len(data.messages)
        while tk_tmp > self.config.llm.session_tokens_windows:
            if len(data.messages) >= 2:
                self._drop_message()
            else:
                break

            tk_tmp: int = get_token(memory_l)
            memory_l = [train_model, *data.messages]
            await asyncio.sleep(
                0
            )  # CPU intensive tasks may cause performance issues, yielding control here
        final_count = len(data.messages)
        logger.debug(
            f"Token count limitation completed, removed {initial_count - final_count} messages"
        )
        logger.debug(f"Final token count: {tk_tmp}")

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Async context manager exit, handle rollback in case of exceptions

        In case of exceptions, restore messages to the state before processing,
        ensuring data consistency.

        Args:
            exc_type: Exception type
            exc_val: Exception value
            exc_tb: Exception traceback
        """
        if exc_type is not None:
            print("An exception occurred, rolling back messages...")
            self.memory.messages = self._copied_messages.messages
            return


class ChatObject:
    """Chat processing object - The minimal unit of chat processing.

    This class is responsible for processing a single chat session, including message receiving,
    context management, model calling, and response sending.
    """

    stream_id: str  # Chat object ID
    timestamp: str  # Timestamp (for LLM)
    time: datetime  # Time
    end_at: datetime | None = None
    data: Memory  # (lateinit) Memory file
    user_input: USER_INPUT
    user_message: Message[USER_INPUT]  # (lateinit) User message
    context_wrap: SendMessageWrap  # (lateinit) Context message
    train: dict[str, str]  # System message
    last_call: datetime  # Last internal function call time
    session_id: str  # Session ID
    response: UniResponse[str, None]  # (lateinit) Response
    preset: ModelPreset  # preset used in this call
    config: AmritaConfig  # config used in this call
    _response_queue: asyncio.Queue[Any]
    _overflow_queue: asyncio.Queue[Any]
    _is_running: bool = False  # Whether it is running
    _is_done: bool = False  # Whether it has completed
    _task: Task[None]
    _err: BaseException | None = None
    _queue_done: bool = False
    _has_consumer: bool = False
    _callback_fun: RESPONSE_CALLBACK_TYPE = None
    _callback_lock: Lock
    __done_marker = object()

    def __init__(
        self,
        train: dict[str, str],
        user_input: USER_INPUT,
        context: Memory | None,
        session_id: str,
        callback: RESPONSE_CALLBACK_TYPE = None,
        config: AmritaConfig | None = None,
        preset: ModelPreset | None = None,
        auto_create_session: bool = False,
        queue_size: int = 25,
        overflow_queue_size: int = 45,
    ) -> None:
        """Initialize chat object

        Args:
            train: Training data (system prompts)
            user_input: Input from the user
            context: Memory context for the session
            session_id: Unique identifier for the session
            callback: Callback function to be called when returning response
            config: Config used for this call
            preset: Preset used for this call
            auto_create_session: Whether to automatically create a session if it does not exist
            queue_size: Maximum number of message chunks to be stored in the queue
            overflow_queue_size: Maximum number of message chunks to be stored in the overflow queue
        """
        sm = SessionsManager()
        if auto_create_session and not sm.is_session_registered(session_id):
            sm.init_session(session_id)
        session = sm.get_session_data(session_id, None)
        self.train = train
        self.data = context or sm.get_session_data(session_id).memory
        self.session_id = session_id
        self.user_input = user_input
        self.user_message = Message(role="user", content=user_input)
        self.timestamp = get_current_datetime_timestamp()
        self.time = datetime.now(utc)
        self.config: AmritaConfig = config or (
            session.config if session else get_config()
        )
        self.last_call = datetime.now(utc)
        self.preset = preset or (
            session.presets.get_default_preset()
            if session
            else PresetManager().get_default_preset()
        )

        # Initialize async queue for streaming responses
        self._response_queue = asyncio.Queue(queue_size)
        self._overflow_queue = asyncio.Queue(overflow_queue_size)
        self.stream_id = uuid4().hex
        self._callback_fun = callback
        self._callback_lock = Lock()

    def get_exception(self) -> BaseException | None:
        """
        Get exceptions that occurred during task execution

        Returns:
            BaseException | None: Returns exception object if an exception occurred during task execution, otherwise returns None
        """
        return self._err

    def is_running(self) -> bool:
        """
        Check if the task is running

        Returns:
            bool: Returns True if the task is running, otherwise returns False
        """
        return self._is_running

    def is_done(self) -> bool:
        """
        Check if the task has completed

        Returns:
            bool: Returns True if the task has completed, otherwise returns False
        """
        return self._is_done

    def terminate(self) -> None:
        """
        Terminate task execution
        Sets the task status to completed and cancels the internal task
        """
        self._is_done = True
        self._is_running = False
        if hasattr(self, "_task") and not self._task.done():
            self._task.cancel()

    def __await__(self):
        if not hasattr(self, "_task"):
            raise RuntimeError("ChatObject not running")
        return self._task.__await__()

    async def __aenter__(self) -> Self:
        if not hasattr(self, "_task"):
            raise RuntimeError("ChatObject not running")
        if self._has_consumer:
            raise RuntimeError("ChatObject already has a consumer")
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.terminate()

    def begin(self) -> Self:
        if not hasattr(self, "_task"):
            logger.debug("Starting chat object task...")
            self._task = asyncio.create_task(self._entry())
        return self

    async def _entry(self) -> None:
        """Call chat object to process messages

        Args:
            event: Message event
            matcher: Matcher
            bot: Bot instance
        """
        if not self._is_running and not self._is_done:
            self.stream_id = uuid4().hex
            logger.debug(f"Starting chat processing, stream ID:{self.stream_id}")

            try:
                self._is_running = True
                self.last_call = datetime.now(utc)
                await chat_manager.add_chat_object(self)

                await self._run()
            finally:
                self._is_running = False
                self._is_done = True
                self.end_at = datetime.now(utc)
                chat_manager.running_chat_object_id2map.pop(self.stream_id, None)
                chat_manager.clean_obj(self.session_id, 1000)  # To avoid memory leaks
                logger.debug("Chat event processing completed")

        else:
            raise RuntimeError(
                f"ChatObject of {self.stream_id} is already running or done"
            )

    async def _run(self):
        """Run chat processing flow

        Execute the main logic of message processing, including getting user information,
        processing message content, managing context length and token limitations,
        and sending responses.
        """
        logger.debug("Starting chat processing flow..")
        data = self.data
        config = self.config

        data.messages.append(self.user_message)

        logger.debug(
            f"Added user message to memory, current message count: {len(data.messages)}"
        )

        self.train["content"] = (
            "<SCHEMA>\n"
            + (
                f"<HIDDEN>{config.cookie.cookie}</HIDDEN>\n"
                if config.cookie.enable_cookie
                else ""
            )
            + "Please participate in the discussion in your own character identity. Try not to use similar phrases when responding to different topics. User's messages are contained within user inputs."
            + "Your character setting is in the <SYSTEM_INSTRUCTIONS> tags, and the summary of previous conversations is in the <SUMMARY> tags."
            + "\n</SCHEMA>\n"
            + "<SYSTEM_INSTRUCTIONS>\n"
            + "\n</SYSTEM_INSTRUCTIONS>"
            + f"\n<SUMMARY>\n{data.abstract if config.llm.enable_memory_abstract else ''}\n</SUMMARY>"
        )
        logger.debug(self.train["content"])

        logger.debug("Starting applying memory limitations..")
        async with MemoryLimiter(self.data, self.train, config=config) as lim:
            await lim.run_enforce()
            abs_usage = lim.usage
            self.data = lim.memory
        logger.debug("Memory limitation application completed")

        send_messages = self._prepare_send_messages()
        self.context_wrap = SendMessageWrap.validate_messages(send_messages)
        logger.debug(
            f"Preparing sending messages completed, message count: {len(send_messages)}"
        )
        response: UniResponse[str, None] = await self._process_chat(send_messages)
        if response.usage and abs_usage:
            response.usage.completion_tokens += abs_usage.completion_tokens
            response.usage.prompt_tokens += abs_usage.prompt_tokens
            response.usage.total_tokens += abs_usage.total_tokens
        self.response = response

        logger.debug("Chat processing completed, preparing to send response")
        await self.set_queue_done()

    def queue_closed(self) -> bool:
        """Check if the response queue is closed

        Returns:
            True if the queue is closed, False otherwise
        """
        return self._queue_done

    async def set_queue_done(self) -> None:
        """Mark the response queue as done by putting the done marker"""
        if not self.queue_closed():
            await self._put_to_queue(self.__done_marker)

    async def _put_to_queue(self, item):
        """Put an item to the queue, using overflow mechanism if primary queue is full

        Args:
            item: Item to put in the queue
        """
        try:
            self._response_queue.put_nowait(item)
        except asyncio.QueueFull:
            try:
                self._overflow_queue.put_nowait(item)
            except asyncio.QueueFull:
                timeout = 5
                while timeout > 0:
                    await asyncio.sleep(1)
                    timeout -= 1
                    try:
                        self._response_queue.put_nowait(item)
                        return
                    except asyncio.QueueFull:
                        try:
                            self._overflow_queue.put_nowait(item)
                            return
                        except asyncio.QueueFull:
                            continue

                # After waiting, if still full, raise an exception
                raise RuntimeError(
                    "Both primary and overflow queues are full after waiting"
                )

    async def yield_response(self, response: RESPONSE_TYPE) -> None:
        """Send chat model response to the queue allowing both str and MessageContent types.

        Args:
            response: Either a string or MessageContent object to be sent to the queue
        """
        if self._callback_fun is not None:
            async with self._callback_lock:
                await self._callback_fun(response)
        else:
            if not self.queue_closed():
                await self._put_to_queue(response)
            else:
                raise RuntimeError("Queue is closed.")

    def set_callback_func(self, func: RESPONSE_CALLBACK_TYPE) -> None:
        """Set a callback function to be executed when a response is yielded.

        Args:
            func (RESPONSE_CALLBACK_TYPE): Function to be executed when a response is yielded

        Raises:
            RuntimeError: If a callback function is already set, raise it.
        """
        if not self._callback_fun:
            self._callback_fun = func
        else:
            raise RuntimeError(
                "The callback function of this chat object has already been set!"
            )

    async def yield_response_iteration(
        self, iterator: AsyncGenerator[RESPONSE_TYPE, None]
    ):
        """Send chat model response to the queue allowing both str and MessageContent types.

        Args:
            iterator: An async generator that yields either strings or MessageContent objects
        """
        async for chunk in iterator:
            await self.yield_response(chunk)

    def get_response_generator(self) -> AsyncGenerator[RESPONSE_TYPE, None]:
        """Return an async generator to iterate over responses from the queue.

        Yields:
            Either a string or MessageContent object from the response queue
        """
        if self._has_consumer or self._callback_fun is not None:
            raise RuntimeError("Response is already being consumed.")
        self._has_consumer = True
        return self._response_generator()

    async def full_response(self) -> str:
        """Return full response from the queue as a single string.

        Returns:
            Complete response string combining all chunks in the queue
        """
        builder = StringIO()
        async for item in self.get_response_generator():
            if isinstance(item, str):
                builder.write(item)
            elif isinstance(item, MessageContent):
                builder.write(str(item.get_content()))
        return builder.getvalue()

    async def _response_generator(self) -> AsyncGenerator[RESPONSE_TYPE]:
        """Internal method to asynchronously yield items from the queue until done marker is reached.

        Yields:
            Items from the response queue until the done marker is encountered
        """
        # Yield from primary queue first
        while True:
            # Check primary queue first
            while not self._response_queue.empty():
                item = await self._response_queue.get()
                self._response_queue.task_done()
                if item is self.__done_marker:
                    return
                yield item

            # If primary queue is empty, check overflow queue
            if not self._overflow_queue.empty():
                item = await self._overflow_queue.get()
                self._overflow_queue.task_done()
                if item is self.__done_marker:
                    return
                yield item
            else:
                if (
                    self.queue_closed()
                    and self._response_queue.empty()
                    and self._overflow_queue.empty()
                ):
                    break
                # Otherwise, wait a bit before checking again
                await asyncio.sleep(0.01)

    async def _process_chat(
        self,
        send_messages: CONTENT_LIST_TYPE,
    ) -> UniResponse[str, None]:
        """Call chat model to generate response and trigger related events.

        Args:
            send_messages: Send message list
            extra_usage: Extra token usage information

        Returns:
            Model response
        """
        self.last_call = datetime.now(utc)

        data = self.data
        logger.debug(
            f"Starting chat processing, sending message count: {len(send_messages)}"
        )

        logger.debug("Triggering matcher functions..")
        messages = self.context_wrap
        chat_event = PreCompletionEvent(
            chat_object=self,
            user_input=self.user_input,
            original_context=messages,  # Use the original context
        )
        await MatcherManager.trigger_event(chat_event, self.config)
        self.data.messages = chat_event.get_context_messages().unwrap(
            exclude_system=True
        )
        send_messages = chat_event.get_context_messages().unwrap()

        logger.debug("Calling chat model..")
        response: UniResponse[str, None] | None = None
        async for chunk in call_completion(
            send_messages, config=self.config, preset=self.preset
        ):
            if isinstance(chunk, str):
                await self.yield_response(StringMessageContent(chunk))
            elif isinstance(chunk, UniResponse):
                response = chunk
            elif isinstance(chunk, MessageContent):
                await self.yield_response(chunk)
        if response is None:
            raise RuntimeError("No final response from chat adapter.")
        self.response = response
        logger.debug("Triggering chat events..")
        chat_event = CompletionEvent(self.user_input, messages, self, response.content)
        await MatcherManager.trigger_event(chat_event, self.config)
        response.content = chat_event.model_response
        data.messages.append(
            Message[str](
                content=response.content,
                role="assistant",
            )
        )
        logger.debug(
            f"Added assistant response to memory, current message count: {len(data.messages)}"
        )

        logger.debug("Chat processing completed")
        return response

    def _prepare_send_messages(
        self,
    ) -> list:
        """Prepare message list to send to the chat model, including system prompt data and context.

        Returns:
            Prepared message list to send
        """
        self.last_call = datetime.now(utc)
        logger.debug("Preparing messages to send..")
        train: Message[str] = Message[str].model_validate(self.train)
        data = self.data
        messages = [train, *copy.deepcopy(data.messages)]
        logger.debug(f"Messages preparation completed, total {len(messages)} messages")
        return messages

    def get_snapshot(self) -> ChatObjectMeta:
        """Get a snapshot of the chat object

        Returns:
            Chat object metadata
        """
        return ChatObjectMeta.model_validate(self, from_attributes=True)


@dataclass
class ChatManager:
    running_chat_object: defaultdict[str, list[ChatObject]] = field(
        default_factory=lambda: defaultdict(list)
    )
    running_chat_object_id2map: dict[str, ChatObjectMeta] = field(default_factory=dict)

    def clean_obj(self, k: str, maxitems: int = 10):
        """
        Clean up running chat objects under the specified key, keeping only the first 10 objects,
        removing any excess unfinished parts

        Args:
            k (tuple[int, bool]): Key value, composed of instance ID and whether it's group chat
            maxitems (int, optional): Maximum number of objects. Defaults to 10.
        """
        objs = self.running_chat_object[k]
        if len(objs) > maxitems:
            dropped_obj = objs[maxitems:]
            objs = [obj for obj in dropped_obj if not obj.is_done()] + objs[:maxitems]
            dropped_obj = [obj for obj in dropped_obj if obj.is_done()]
            for obj in dropped_obj:
                self.running_chat_object_id2map.pop(obj.stream_id, None)
            self.running_chat_object[k] = objs

    def get_all_objs(self) -> list[ChatObjectMeta]:
        """
        Get all running chat object metadata

        Returns:
            list[ChatObjectMeta]: List of all running chat object metadata
        """
        return list(self.running_chat_object_id2map.values())

    def get_objs(self, session_id: str) -> list[ChatObject]:
        """
        Get the corresponding list of chat objects based on the session ID

        Args:
            session_id (str): User session ID

        Returns:
            list[ChatObject]: List of chat objects
        """
        return self.running_chat_object[session_id]

    async def clean_chat_objects(self, maxitems: int = 10) -> None:
        """
        Asynchronously clean up all running chat objects, limiting the number of objects for each key to no more than 10
        """
        async with LOCK:
            for key in self.running_chat_object.keys():
                self.clean_obj(key, maxitems)

    async def add_chat_object(self, chat_object: ChatObject) -> None:
        """
        Add a new chat object to the running list

        Args:
            chat_object (ChatObject): Chat object instance
        """
        async with LOCK:
            meta: ChatObjectMeta = chat_object.get_snapshot()
            self.running_chat_object_id2map[chat_object.stream_id] = meta
            key = chat_object.session_id
            self.running_chat_object[key].insert(0, chat_object)
            self.clean_obj(key)


chat_manager = ChatManager()



from __future__ import annotations

from datetime import datetime
from typing import Any

import pytz


def remove_think_tag(text: str) -> str:
    """Remove the first occurrence of think tag

    Args:
        text (str): Parameter to process

    Returns:
        str: Processed text
    """

    start_tag = "<think>"
    end_tag = "</think>"

    # Find the position of the first start tag
    start_idx = text.find(start_tag)
    if start_idx == -1:
        return text  # No start tag found, return original text

    # Find the position of the end tag after the start tag
    end_idx = text.find(end_tag, start_idx + len(start_tag))
    if end_idx == -1:
        return text  # No corresponding end tag found, return original text

    # Calculate the end position of the end tag
    end_of_end_tag = end_idx + len(end_tag)

    # Concatenate text after removing the tag
    text_new = text[:start_idx] + text[end_of_end_tag:]
    while text_new.startswith("\n"):
        text_new = text_new[1:]
    return text_new


def format_datetime_timestamp(time: float) -> str:
    """Format timestamp to date, weekday and time string"""
    now = datetime.fromtimestamp(time)
    formatted_date = now.strftime("%Y-%m-%d")
    formatted_weekday = now.strftime("%A")
    formatted_time = now.strftime("%I:%M:%S %p")
    return f"[{formatted_date} {formatted_weekday} {formatted_time}]"


def split_list(lst: list, threshold: int) -> list[Any]:
    """Split list into multiple sublists, each sublist length does not exceed threshold"""
    if len(lst) <= threshold:
        return [lst]
    return [lst[i : i + threshold] for i in range(0, len(lst), threshold)]


def get_current_datetime_timestamp(utc_time: None | datetime = None):
    """Get current time and format as date, weekday and time string"""
    utc_time = utc_time or datetime.now(pytz.utc)
    asia_shanghai = pytz.timezone("Asia/Shanghai")
    now = utc_time.astimezone(asia_shanghai)
    formatted_date = now.strftime("%Y-%m-%d")
    formatted_weekday = now.strftime("%A")
    formatted_time = now.strftime("%H:%M:%S")
    return f"[{formatted_date} {formatted_weekday} {formatted_time}]"



import asyncio

import pytest

from amrita_core.chatmanager import ChatObject, MemoryLimiter, chat_manager
from amrita_core.config import AmritaConfig
from amrita_core.sessions import SessionsManager
from amrita_core.types import (
    CONTENT_LIST_TYPE_ITEM,
    MemoryModel,
    Message,
    ModelPreset,
)


class TestChatObject:
    """Test ChatObject class functionality"""

    def setup_method(self):
        """Clean up state before each test method"""
        # Reset SessionsManager
        SessionsManager._instance = None

    @pytest.mark.asyncio
    async def test_chat_object_initialization(self):
        """Test ChatObject initialization"""
        session_id = "test-session-123"
        sm = SessionsManager()
        sm.init_session(session_id)

        train = {"role": "system", "content": "system message"}
        user_input = "hello"
        context = MemoryModel()

        # Create a default preset to avoid the random choice error
        default_preset = ModelPreset(
            model="gpt-3.5-turbo", name="test-default", api_key="fake-key"
        )

        chat_obj = ChatObject(
            train=train,
            user_input=user_input,
            context=context,
            session_id=session_id,
            preset=default_preset,  # Pass a preset to avoid the error
        )

        assert chat_obj.session_id == session_id
        assert chat_obj.user_input == user_input
        assert chat_obj.train == train

    @pytest.mark.asyncio
    async def test_chat_object_run_flow(self):
        """Test ChatObject run flow"""
        session_id = "test-session-456"
        sm = SessionsManager()
        sm.init_session(session_id)

        train = {"role": "system", "content": "system message"}
        user_input = "test input"
        context = MemoryModel()

        # Create a Mock configuration to avoid actual LLM calls
        config = AmritaConfig()
        config.llm.enable_tokens_limit = False
        config.llm.enable_memory_abstract = False

        # Create a default preset to avoid the random choice error
        default_preset = ModelPreset(
            model="gpt-3.5-turbo", name="test-default", api_key="fake-key"
        )

        chat_obj = ChatObject(
            train=train,
            user_input=user_input,
            context=context,
            session_id=session_id,
            config=config,
            preset=default_preset,  # Pass a preset to avoid the error
        )

        # Test if it can start running
        chat_obj.begin()
        assert chat_obj.is_running() or not chat_obj.is_done()

        # Terminate object
        chat_obj.terminate()

    @pytest.mark.asyncio
    async def test_response_generator(self):
        """Test response generator"""
        session_id = "test-session-789"
        sm = SessionsManager()
        sm.init_session(session_id)

        train = {"role": "system", "content": "system message"}
        user_input = "test input"
        context = MemoryModel()

        config = AmritaConfig()
        config.llm.enable_tokens_limit = False

        # Create a default preset to avoid the random choice error
        default_preset = ModelPreset(
            model="gpt-3.5-turbo", name="test-default", api_key="fake-key"
        )

        chat_obj = ChatObject(
            train=train,
            user_input=user_input,
            context=context,
            session_id=session_id,
            config=config,
            preset=default_preset,  # Pass a preset to avoid the error
        )

        # Add some mock responses to queue
        await chat_obj._put_to_queue("Hello")
        await chat_obj._put_to_queue("World")
        await chat_obj.set_queue_done()

        # Get response generator and verify
        gen = chat_obj.get_response_generator()
        responses = [resp async for resp in gen]

        assert len(responses) == 2
        assert "Hello" in responses
        assert "World" in responses


class TestMemoryLimiter:
    """Test MemoryLimiter class functionality"""

    @pytest.mark.asyncio
    async def test_memory_limiter_initialization(self):
        """Test MemoryLimiter initialization"""
        memory = MemoryModel()
        train = {"role": "system", "content": "system prompt"}
        config = AmritaConfig()

        limiter = MemoryLimiter(memory, train, config)

        assert limiter.memory == memory
        assert limiter.config == config
        assert limiter._train == train

    @pytest.mark.asyncio
    async def test_memory_limiter_context_manager(self):
        """Test MemoryLimiter context manager"""
        memory = MemoryModel()
        train = {"role": "system", "content": "system prompt"}
        config = AmritaConfig()

        async with MemoryLimiter(memory, train, config) as lim:
            # Verify internal attributes are initialized after entering context manager
            assert hasattr(lim, "_dropped_messages")
            assert hasattr(lim, "_copied_messages")
            assert isinstance(lim._dropped_messages, list)
            assert lim._copied_messages == memory

    @pytest.mark.asyncio
    async def test_memory_length_limit(self):
        """Test memory length limit functionality"""
        # Create memory with multiple messages
        messages: list[CONTENT_LIST_TYPE_ITEM] = [
            Message(role="user", content=f"message {i}")
            for i in range(20)  # Create enough messages to trigger limit
        ]
        memory = MemoryModel(messages=messages)
        train = {"role": "system", "content": "system prompt"}

        # Set a small memory limit
        config = AmritaConfig()
        config.llm.memory_lenth_limit = 5  # Limit to max 5 messages

        async with MemoryLimiter(memory, train, config) as lim:
            await lim._limit_length()

            # Verify message count is limited to the configured limit
            assert len(lim.memory.messages) <= config.llm.memory_lenth_limit


class TestChatManager:
    """Test ChatManager class functionality"""

    def setup_method(self):
        """Clean up chat_manager state before each test method"""
        chat_manager.running_chat_object.clear()
        chat_manager.running_chat_object_id2map.clear()

    @pytest.mark.asyncio
    async def test_add_and_get_chat_objects(self):
        """Test add and get chat objects"""
        session_id = "test-session-add-get"
        sm = SessionsManager()
        sm.init_session(session_id)

        train = {"role": "system", "content": "system message"}
        user_input = "test input"
        context = MemoryModel()

        # Create a default preset to avoid the random choice error
        default_preset = ModelPreset(
            model="gpt-3.5-turbo", name="test-default", api_key="fake-key"
        )

        # Create chat object
        chat_obj = ChatObject(
            train=train,
            user_input=user_input,
            context=context,
            session_id=session_id,
            preset=default_preset,  # Pass a preset to avoid the error
        )

        # Add to manager
        await chat_manager.add_chat_object(chat_obj)

        # Verify object is added
        objs = chat_manager.get_objs(session_id)
        assert len(objs) == 1
        assert objs[0] == chat_obj

    @pytest.mark.asyncio
    async def test_get_all_objs(self):
        """Test get all objects metadata"""
        session_id = "test-session-all-objs"
        sm = SessionsManager()
        sm.init_session(session_id)

        train = {"role": "system", "content": "system message"}
        user_input = "test input"
        context = MemoryModel()

        # Create a default preset to avoid the random choice error
        default_preset = ModelPreset(
            model="gpt-3.5-turbo", name="test-default", api_key="fake-key"
        )

        # Create and add chat object
        chat_obj = ChatObject(
            train=train,
            user_input=user_input,
            context=context,
            session_id=session_id,
            preset=default_preset,  # Pass a preset to avoid the error
        )

        await chat_manager.add_chat_object(chat_obj)

        # Get all object metadata
        all_objs = chat_manager.get_all_objs()

        # Verify at least one object metadata exists
        assert len(all_objs) >= 1
        assert chat_obj.stream_id in [meta.stream_id for meta in all_objs]

    @pytest.mark.asyncio
    async def test_clean_chat_objects(self):
        """Test clean all chat objects"""
        session_id = "test-session-clean-all"
        sm = SessionsManager()
        sm.init_session(session_id)

        # Add multiple objects
        for i in range(3):
            train = {"role": "system", "content": f"system message {i}"}
            user_input = f"test input {i}"
            context = MemoryModel()

            # Create a default preset to avoid the random choice error
            default_preset = ModelPreset(
                model="gpt-3.5-turbo", name=f"test-{i}", api_key="fake-key"
            )

            chat_obj = ChatObject(
                train=train,
                user_input=user_input,
                context=context,
                session_id=session_id,
                preset=default_preset,  # Pass a preset to avoid the error
            )
            chat_obj.terminate()

            await chat_manager.add_chat_object(chat_obj)

        # Clean all chat objects
        await chat_manager.clean_chat_objects(maxitems=2)

        # Verify object count is limited
        objs = chat_manager.get_objs(session_id)
        assert len(objs) <= 2


@pytest.mark.asyncio
async def test_concurrent_chat_objects():
    """Test concurrent chat objects"""
    sm = SessionsManager()

    # Create multiple sessions
    session_ids = []
    for i in range(3):
        session_id = sm.new_session()
        session_ids.append(session_id)

    # Concurrently create multiple chat objects
    async def create_chat_obj(session_id):
        train = {"role": "system", "content": f"system message for {session_id}"}
        user_input = f"test input for {session_id}"
        context = MemoryModel()

        # Create a default preset to avoid the random choice error
        default_preset = ModelPreset(
            model="gpt-3.5-turbo", name=f"test-{session_id[:8]}", api_key="fake-key"
        )

        chat_obj = ChatObject(
            train=train,
            user_input=user_input,
            context=context,
            session_id=session_id,
            preset=default_preset,  # Pass a preset to avoid the error
        )

        await chat_manager.add_chat_object(chat_obj)
        return chat_obj

    # Concurrently create chat objects
    chat_objects = await asyncio.gather(*[create_chat_obj(sid) for sid in session_ids])

    # Verify all objects are created and added
    for i, sid in enumerate(session_ids):
        objs = chat_manager.get_objs(sid)
        assert len(objs) >= 1

        # Check if corresponding chat object exists
        assert any(obj.session_id == sid for obj in objs)

    # Clean up all objects
    for obj in chat_objects:
        obj.terminate()



from .event import EventTypeEnum
from .matcher import Matcher


def on_completion(priority: int = 10, block: bool = True):
    return on_event(EventTypeEnum.COMPLETION, priority, block)


def on_precompletion(priority: int = 10, block: bool = True):
    return on_event(EventTypeEnum.BEFORE_COMPLETION, priority, block)


def on_event(event_type: EventTypeEnum, priority: int = 10, block: bool = True):
    return Matcher(event_type, priority, block)



from __future__ import annotations

import typing
from collections.abc import AsyncGenerator, Generator

from pydantic import ValidationError

from amrita_core.preset import PresetManager

from .config import AmritaConfig, get_config
from .logging import debug_log
from .protocol import (
    AdapterManager,
    ModelAdapter,
)
from .tokenizer import hybrid_token_count
from .tools.models import ToolChoice
from .types import (
    CONTENT_LIST_TYPE,
    Message,
    ModelPreset,
    ToolCall,
    ToolResult,
    UniResponse,
    UniResponseUsage,
)

T = typing.TypeVar("T")


def text_generator(
    memory: CONTENT_LIST_TYPE, split_role: bool = False
) -> Generator[str, None, str]:
    """Generator that yields text content from a list of messages.

    Args:
        memory: List of message objects containing content
        split_role: Whether to prepend role-specific prefixes to content

    Yields:
        Individual text strings from the message content
    """
    memory_l = [(i.model_dump() if hasattr(i, "model_dump") else i) for i in memory]
    role_map = {
        "assistant": "<BOT's response>",
        "user": "<User's query>",
        "tool": "<Tool call>",
    }
    for st in memory_l:
        if st["content"] is None:
            continue
        if isinstance(st["content"], str):
            yield (
                st["content"]
                if not split_role
                else role_map.get(st["role"], "") + st["content"]
            )
        else:
            for s in st["content"]:
                if s["type"] == "text" and s.get("text") is not None:
                    yield (
                        s["text"]
                        if not split_role
                        else role_map.get(st["role"], "") + s["text"]
                    )
    return ""


async def get_tokens(
    memory: CONTENT_LIST_TYPE,
    response: UniResponse[str, None],
    config: AmritaConfig | None = None,
) -> UniResponseUsage[int]:
    """Calculate token counts for messages and response

    Args:
        memory: Message history list
        response: Model response
        config: Optional configuration to use (uses default if not provided)

    Returns:
        Object containing token usage information
    """
    if (
        response.usage is not None
        and response.usage.total_tokens is not None
        and response.usage.completion_tokens is not None
        and response.usage.prompt_tokens is not None
    ):
        return response.usage
    config = config or get_config()
    it = hybrid_token_count(
        "".join(list(text_generator(memory))),
        config.llm.tokens_count_mode,
    )

    ot = hybrid_token_count(response.content)
    return UniResponseUsage(
        prompt_tokens=it, total_tokens=it + ot, completion_tokens=ot
    )


def _validate_msg_list(
    messages: CONTENT_LIST_TYPE,
) -> CONTENT_LIST_TYPE:
    """Validate a list of message dictionaries and convert them to Message objects.

    Args:
        messages: List of message dictionaries or Message objects

    Returns:
        List of validated Message objects

    Raises:
        ValueError: If a message dictionary is invalid
    """
    validated_messages = []
    for msg in messages:
        if isinstance(msg, dict):
            # Ensure message has role field
            if "role" not in msg:
                raise ValueError("Message dictionary is missing 'role' field")
            try:
                validated_msg = (
                    Message.model_validate(msg)
                    if msg["role"] != "tool"
                    else ToolResult.model_validate(msg)
                )
            except ValidationError as e:
                raise ValueError(f"Invalid message format: {e}")
            validated_messages.append(validated_msg)
        else:
            validated_messages.append(msg)
    return validated_messages


async def _call_with_reflection(
    preset: ModelPreset,
    call_func: typing.Callable[..., typing.Awaitable[T]],
    config: AmritaConfig,
    *args,
    **kwargs,
) -> T:
    """Internal helper to call an adapter function with reflection and logging.

    Args:
        preset: Model preset to use for the call
        call_func: Async function to call on the adapter
        config: Configuration to pass to the adapter
        *args: Arguments to pass to the call function
        **kwargs: Keyword arguments to pass to the call function

    Returns:
        Result of the call function
    """
    adapter_class = AdapterManager().safe_get_adapter(preset.protocol)
    if adapter_class:
        debug_log(
            f"Using adapter {adapter_class.__name__} to handle protocol {preset.protocol}"
        )
    else:
        raise ValueError(f"Undefined protocol adapter: {preset.protocol}")

    debug_log(f"Getting chat for {preset.model}")
    debug_log(f"Preset: {preset.name}")
    debug_log(f"Key: {preset.api_key[:7]}...")
    debug_log(f"Protocol: {preset.protocol}")
    debug_log(f"API URL: {preset.base_url}")
    debug_log(f"Model: {preset.model}")
    adapter = adapter_class(preset, config)
    return await call_func(adapter, *args, **kwargs)


async def tools_caller(
    messages: CONTENT_LIST_TYPE,
    tools: list,
    preset: ModelPreset | None = None,
    tool_choice: ToolChoice | None = None,
    config: AmritaConfig | None = None,
) -> UniResponse[None, list[ToolCall] | None]:
    """Call tools using the specified model preset.

    Args:
        messages: List of messages to send to the model
        tools: List of available tools
        preset: Model preset to use (uses default if not provided)
        tool_choice: How to select tools (uses default if not provided)
        config: Configuration to use (uses default if not provided)

    Returns:
        Response containing tool calls or None
    """
    config = config or get_config()

    async def _call_tools(
        adapter: ModelAdapter,
        messages: CONTENT_LIST_TYPE,
        tools,
        tool_choice,
    ):
        return await adapter.call_tools(messages, tools, tool_choice)

    preset = preset or PresetManager().get_default_preset()
    return await _call_with_reflection(
        preset, _call_tools, config, messages, tools, tool_choice
    )


async def call_completion(
    messages: CONTENT_LIST_TYPE,
    preset: ModelPreset | None = None,
    config: AmritaConfig | None = None,
) -> AsyncGenerator[str | UniResponse[str, None], None]:
    """Get chat response from the model.

    Args:
        messages: List of messages to send to the model
        preset: Model preset to use (uses default if not provided)
        config: Configuration to use (uses default if not provided)

    Yields:
        Individual response parts as strings or UniResponse objects
    """
    messages = _validate_msg_list(messages)
    preset = preset or PresetManager().get_default_preset()
    config = config or get_config()

    async def _call_api(adapter: ModelAdapter, messages: CONTENT_LIST_TYPE):
        async def inner():
            async for i in adapter.call_api([(i.model_dump()) for i in messages]):
                yield i

        return inner

    # Call adapter to get chat response
    response = await _call_with_reflection(preset, _call_api, config, messages)

    async for i in response():
        yield i


async def get_last_response(
    generator: AsyncGenerator[str | UniResponse[str, None], None],
) -> UniResponse[str, None]:
    """Extract the last UniResponse from a response generator.

    Args:
        generator: Async generator yielding response parts

    Returns:
        The last UniResponse object from the generator

    Raises:
        RuntimeError: If no response is found in the generator
    """
    ls: list[UniResponse[str, None]] = [
        i async for i in generator if isinstance(i, UniResponse)
    ]
    if len(ls) == 0:
        raise RuntimeError("No response found in generator.")
    return ls[-1]



class MatcherException(Exception):
    """Base exception for Matcher."""


class BlockException(MatcherException):
    pass


class CancelException(MatcherException):
    pass


class PassException(MatcherException):
    pass



import asyncio

import pytest

from amrita_core.types import (
    CONTENT_LIST_TYPE,
    Function,
    MemoryModel,
    Message,
    ModelConfig,
    ModelPreset,
    SendMessageWrap,
    ToolCall,
    UniResponse,
)


class TestModelConfig:
    """Test ModelConfig class functionality"""

    def test_default_values(self):
        """Test default values"""
        config = ModelConfig()
        assert config.top_k
        assert config.top_p
        assert config.temperature
        assert config.stream is False
        assert config.multimodal is False

    def test_custom_values(self):
        """Test custom values"""
        config = ModelConfig(top_k=100, temperature=0.8, stream=True)
        assert config.top_k == 100
        assert config.temperature == 0.8
        assert config.stream is True


class TestModelPreset:
    """Test ModelPreset class functionality"""

    def test_default_values(self):
        """Test default values"""
        preset = ModelPreset()
        assert preset.model == ""
        assert preset.name == "default"
        assert preset.base_url == ""
        assert preset.api_key == ""
        assert preset.protocol == "__main__"

    def test_load_save(self, tmp_path):
        """Test save and load functionality"""
        # Create temporary file path
        temp_file = tmp_path / "preset.json"

        # Create a preset and save it
        preset = ModelPreset(
            model="gpt-3.5-turbo", name="test_preset", api_key="test_key"
        )
        preset.save(temp_file)

        # Verify file exists
        assert temp_file.exists()

        # Load preset from file
        loaded_preset = ModelPreset.load(temp_file)

        # Verify loaded preset values
        assert loaded_preset.model == "gpt-3.5-turbo"
        assert loaded_preset.name == "test_preset"
        assert loaded_preset.api_key == "test_key"

    def test_extra_field(self):
        """Test extra field functionality"""
        preset = ModelPreset(extra={"custom_param": "value"})
        assert preset.extra["custom_param"] == "value"


class TestMessage:
    """Test Message class functionality"""

    def test_text_message(self):
        """Test text message"""
        msg = Message(role="user", content="Hello")
        assert msg.role == "user"
        assert msg.content == "Hello"


class TestSendMessageWrap:
    """Test SendMessageWrap class functionality"""

    def test_initialization(self):
        """Test initialization functionality"""
        train_msg = {"role": "system", "content": "system message"}
        memory: CONTENT_LIST_TYPE = [Message(role="user", content="previous message")]
        user_query = Message(role="user", content="current query")

        wrapper = SendMessageWrap(train_msg, memory, user_query)

        assert wrapper.train.content == "system message"
        assert wrapper.user_query.content == "current query"
        assert len(wrapper.memory) == 1

    def test_validation(self):
        """Test message validation functionality"""
        system_msg = Message(role="system", content="system message")
        user_msg = Message(role="user", content="user message")
        assistant_msg = Message(role="assistant", content="assistant message")

        messages: CONTENT_LIST_TYPE = [system_msg, user_msg, assistant_msg, user_msg]
        wrapper = SendMessageWrap.validate_messages(messages)

        assert wrapper.train.content == "system message"
        assert len(wrapper.memory) == 2
        assert wrapper.user_query.content == "user message"

    def test_iterable_behavior(self):
        """Test iterable behavior"""
        train_msg = {"role": "system", "content": "system message"}
        memory: CONTENT_LIST_TYPE = [
            Message(role="assistant", content="assistant message")
        ]
        user_query = Message(role="user", content="user query")

        wrapper = SendMessageWrap(train_msg, memory, user_query)

        items = list(wrapper)
        assert len(items) == 3  # train + memory + user_query
        assert items[0].content == "system message"
        assert items[1].content == "assistant message"
        assert items[2].content == "user query"


class TestUniResponse:
    """Test UniResponse class functionality"""

    def test_basic_response(self):
        """Test basic response"""
        response = UniResponse(content="Hello world", tool_calls=None, role="assistant")
        assert response.content == "Hello world"
        assert response.tool_calls is None
        assert response.role == "assistant"

    def test_response_with_tool_calls(self):
        """Test response with tool calls"""
        func = Function(arguments='{"x": 5}', name="add")
        tool_call = ToolCall(id="call_1", function=func)

        response = UniResponse(content="Calculating...", tool_calls=[tool_call])
        assert response.content == "Calculating..."
        assert len(response.tool_calls) == 1
        assert response.tool_calls[0].id == "call_1"


class TestMemoryModel:
    """Test MemoryModel class functionality"""

    def test_default_values(self):
        """Test default values"""
        memory = MemoryModel()
        assert memory.messages == []
        assert memory.abstract == ""
        assert memory.time > 0  # Should be current timestamp

    def test_with_messages(self):
        """Test with messages"""
        msg = Message(role="user", content="test message")
        memory = MemoryModel(messages=[msg])

        assert len(memory.messages) == 1
        assert memory.messages[0].role == "user"
        assert memory.messages[0].content == "test message"


@pytest.mark.asyncio
async def test_async_example():
    """Async test example"""
    # Simulate an async operation
    await asyncio.sleep(0.01)
    assert True



from collections.abc import AsyncGenerator, Iterable

import openai
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from openai.types.chat.chat_completion_named_tool_choice_param import (
    ChatCompletionNamedToolChoiceParam,
)
from openai.types.chat.chat_completion_named_tool_choice_param import (
    Function as OPENAI_Function,
)
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionToolChoiceOptionParam,
)
from typing_extensions import override

from amrita_core.config import AmritaConfig
from amrita_core.logging import debug_log
from amrita_core.protocol import ModelAdapter
from amrita_core.tools.models import ToolChoice, ToolFunctionSchema
from amrita_core.types import (
    ModelConfig,
    ModelPreset,
    ToolCall,
    UniResponse,
    UniResponseUsage,
)


class OpenAIAdapter(ModelAdapter):
    """OpenAI Protocol Adapter"""

    __override__ = True

    @override
    async def call_api(
        self, messages: Iterable[ChatCompletionMessageParam]
    ) -> AsyncGenerator[str | UniResponse[str, None], None]:
        """Call OpenAI API to get chat responses"""
        preset: ModelPreset = self.preset
        preset_config: ModelConfig = preset.config
        config: AmritaConfig = self.config
        client = openai.AsyncOpenAI(
            base_url=preset.base_url,
            api_key=preset.api_key,
            timeout=config.llm.llm_timeout,
            max_retries=config.llm.max_retries,
        )
        completion: ChatCompletion | openai.AsyncStream[ChatCompletionChunk] | None = (
            None
        )
        if stream := preset.config.stream:
            completion = await client.chat.completions.create(
                model=preset.model,
                messages=messages,
                max_tokens=config.llm.max_tokens,
                top_p=preset_config.top_p,
                temperature=preset_config.temperature,
                stream=stream,
                stream_options={"include_usage": True},
            )
        else:
            completion = await client.chat.completions.create(
                model=preset.model,
                messages=messages,
                max_tokens=config.llm.max_tokens,
                top_p=preset_config.top_p,
                temperature=preset_config.temperature,
                stream=False,
            )
        response: str = ""
        uni_usage = None
        # Process streaming response
        if self.preset.config.stream and isinstance(completion, openai.AsyncStream):
            async for chunk in completion:
                try:
                    if chunk.usage:
                        uni_usage = UniResponseUsage.model_validate(
                            chunk.usage, from_attributes=True
                        )
                    if chunk.choices[0].delta.content is not None:
                        response += chunk.choices[0].delta.content
                        yield chunk.choices[0].delta.content
                        debug_log(chunk.choices[0].delta.content)
                except IndexError:
                    break
        else:
            debug_log(response)
            if isinstance(completion, ChatCompletion):
                response = (
                    completion.choices[0].message.content
                    if completion.choices[0].message.content is not None
                    else ""
                )
                yield response
                if completion.usage:
                    uni_usage = UniResponseUsage.model_validate(
                        completion.usage, from_attributes=True
                    )
            else:
                raise RuntimeError("Received unexpected response type")
        uni_response = UniResponse(
            role="assistant",
            content=response,
            usage=uni_usage,
            tool_calls=None,
        )
        yield uni_response

    @override
    async def call_tools(
        self,
        messages: Iterable,
        tools: list,
        tool_choice: ToolChoice | None = None,
    ) -> UniResponse[None, list[ToolCall] | None]:
        if not tool_choice:
            choice: ChatCompletionToolChoiceOptionParam = "auto"
        elif isinstance(tool_choice, ToolFunctionSchema):
            choice = ChatCompletionNamedToolChoiceParam(
                function=OPENAI_Function(name=tool_choice.function.name),
                type=tool_choice.type,
            )
        else:
            choice = tool_choice
        config: AmritaConfig = self.config
        preset: ModelPreset = self.preset
        preset_config: ModelConfig = preset.config
        base_url: str = preset.base_url
        key: str = preset.api_key
        model: str = preset.model
        client = openai.AsyncOpenAI(
            base_url=base_url,
            api_key=key,
            timeout=config.llm.llm_timeout,
        )
        completion: ChatCompletion = await client.chat.completions.create(
            model=model,
            messages=messages,
            stream=False,
            tool_choice=choice,
            tools=tools,
            top_p=preset_config.top_p,
            temperature=preset_config.temperature,
        )
        msg = completion.choices[0].message
        return UniResponse(
            role="assistant",
            tool_calls=(
                [
                    ToolCall.model_validate(i, from_attributes=True)
                    for i in msg.tool_calls
                ]
                if msg.tool_calls
                else None
            ),
            content=None,
        )

    @staticmethod
    def get_adapter_protocol() -> tuple[str, str]:
        return "openai", "__main__"



import asyncio

import pytest

from amrita_core import init
from amrita_core.chatmanager import ChatObject, chat_manager
from amrita_core.config import AmritaConfig
from amrita_core.sessions import SessionsManager
from amrita_core.types import MemoryModel, Message, ModelPreset


class TestIntegration:
    """Test integration between various modules"""

    def setup_method(self):
        """Reset state before each test method"""
        SessionsManager._instance = None
        chat_manager.running_chat_object.clear()
        chat_manager.running_chat_object_id2map.clear()

    @pytest.mark.asyncio
    async def test_session_creation_and_chat_flow(self):
        """Test session creation and chat flow integration"""
        # Initialize AmritaCore
        init()

        # Create a new session
        sm = SessionsManager()
        session_id = sm.new_session()

        # Verify session is created
        assert sm.is_session_registered(session_id)

        # Get session data
        session_data = sm.get_session_data(session_id)
        config = session_data.config
        tools = session_data.tools
        presets = session_data.presets

        assert isinstance(config, AmritaConfig)
        assert tools is not None
        assert presets is not None

        # Create a default preset to avoid the random choice error
        default_preset = ModelPreset(
            model="gpt-3.5-turbo", name="test-default", api_key="fake-key"
        )
        presets.set_default_preset(default_preset)

        # Create a simple chat object
        train = {"role": "system", "content": "You are a helpful assistant."}
        user_input = "Hello, how are you?"
        context = MemoryModel()

        # Create chat object using session config
        chat_obj = ChatObject(
            train=train,
            user_input=user_input,
            context=context,
            session_id=session_id,
            config=config,
            preset=default_preset,  # Pass a preset to avoid the error
        )

        # Add to chat manager
        await chat_manager.add_chat_object(chat_obj)

        # Verify chat object is added
        session_objects = chat_manager.get_objs(session_id)
        assert len(session_objects) >= 1

        # Get all running object metadata
        all_objs = chat_manager.get_all_objs()
        assert len([obj for obj in all_objs if obj.session_id == session_id]) >= 1

    @pytest.mark.asyncio
    async def test_multiple_sessions_with_different_configs(self):
        """Test multiple sessions with different configurations"""
        # Create multiple sessions
        sm = SessionsManager()

        configs = []
        session_ids = []

        for i in range(3):
            # Create custom config
            config = AmritaConfig()
            config.llm.max_retries = 3 + i  # Different retry settings

            # Create session
            session_id = sm.new_session()
            session_ids.append(session_id)

            # Get session data and set config
            session_data = sm.get_session_data(session_id)
            session_data.config = config

            # Save config reference for later verification
            configs.append(config)

        # Verify all sessions are registered
        assert len(sm.get_registered_sessions()) == 3

        # Verify each session has the correct config
        for i, sid in enumerate(session_ids):
            session_data = sm.get_session_data(sid)
            session_config = session_data.config
            assert session_config.llm.max_retries == configs[i].llm.max_retries

    @pytest.mark.asyncio
    async def test_memory_interaction_with_chat_object(self):
        """Test memory interaction with chat object"""
        # Create session
        sm = SessionsManager()
        session_id = sm.new_session()

        # Create memory with initial messages
        initial_memory = MemoryModel()
        initial_message = Message(role="assistant", content="Previous conversation")
        initial_memory.messages.append(initial_message)

        # Create a default preset to avoid the random choice error
        default_preset = ModelPreset(
            model="gpt-3.5-turbo", name="test-default", api_key="fake-key"
        )

        # Create chat object
        train = {"role": "system", "content": "You are a helpful assistant."}
        user_input = "What did we talk about before?"

        chat_obj = ChatObject(
            train=train,
            user_input=user_input,
            context=initial_memory,
            session_id=session_id,
            preset=default_preset,  # Pass a preset to avoid the error
        )

        # Verify chat object's user input
        assert chat_obj.user_input == user_input

        # Verify chat object's memory
        assert len(chat_obj.data.messages) == 1  # Should contain initial message
        assert chat_obj.data.messages[0].content == "Previous conversation"

    @pytest.mark.asyncio
    async def test_preset_integration(self):
        """Test preset integration with chat object"""
        # Create session
        sm = SessionsManager()
        session_id = sm.new_session()

        # Get session's preset manager
        session_data = sm.get_session_data(session_id)
        presets_manager = session_data.presets

        # Create a custom preset
        custom_preset = ModelPreset(
            model="gpt-4-test",
            name="integration-test-preset",
            base_url="https://api.openai.com/v1",  # Using a valid field
        )

        # Set preset as default
        presets_manager.set_default_preset(custom_preset)

        # Create chat object using this preset
        train = {"role": "system", "content": "You are a helpful assistant."}
        user_input = "Test preset integration"
        context = MemoryModel()

        chat_obj = ChatObject(
            train=train,
            user_input=user_input,
            context=context,
            session_id=session_id,
            preset=custom_preset,
        )

        # Verify chat object uses correct preset
        assert chat_obj.preset.name == "integration-test-preset"
        assert chat_obj.preset.model == "gpt-4-test"


@pytest.mark.asyncio
async def test_full_workflow():
    """Test full workflow"""
    # Initialize
    init()

    # Create session
    sm = SessionsManager()
    session_id = sm.new_session()

    # Prepare chat parameters
    train = {"role": "system", "content": "You are a helpful assistant."}
    user_input = "Hello, let's test the full workflow!"
    context = MemoryModel()

    # Create a default preset to avoid the random choice error
    default_preset = ModelPreset(
        model="gpt-3.5-turbo", name="test-default", api_key="fake-key"
    )

    # Create and start chat object
    chat_obj = ChatObject(
        train=train,
        user_input=user_input,
        context=context,
        session_id=session_id,
        preset=default_preset,  # Pass a preset to avoid the error
    ).begin()

    # Add to manager
    await chat_manager.add_chat_object(chat_obj)

    # Wait a bit for object processing
    await asyncio.sleep(0.05)

    # Verify object status
    assert chat_obj.session_id == session_id

    # Verify manager has object
    session_objects = chat_manager.get_objs(session_id)
    assert len(session_objects) >= 1

    # Cleanup
    chat_obj.terminate()


@pytest.mark.asyncio
async def test_concurrent_sessions_and_objects():
    """Test concurrent sessions and objects"""
    # Initialize
    init()

    # Create multiple sessions
    sm = SessionsManager()
    session_ids = [sm.new_session() for _ in range(3)]

    # Concurrently create chat objects
    async def create_and_manage_chat(session_id):
        train = {"role": "system", "content": f"System for session {session_id}"}
        user_input = f"Test message for {session_id}"
        context = MemoryModel()

        # Create a default preset to avoid the random choice error
        default_preset = ModelPreset(
            model="gpt-3.5-turbo", name=f"test-{session_id[:8]}", api_key="fake-key"
        )

        chat_obj = ChatObject(
            train=train,
            user_input=user_input,
            context=context,
            session_id=session_id,
            preset=default_preset,  # Pass a preset to avoid the error
        ).begin()

        await chat_manager.add_chat_object(chat_obj)
        return chat_obj

    # Concurrently create multiple chat objects
    chat_objects = await asyncio.gather(
        *[create_and_manage_chat(sid) for sid in session_ids]
    )

    # Verify all objects are correctly created and managed
    for i, sid in enumerate(session_ids):
        session_objects = chat_manager.get_objs(sid)
        assert len(session_objects) >= 1

        # Check if corresponding chat object exists
        assert any(obj.session_id == sid for obj in session_objects)

    # Clean up all objects
    for obj in chat_objects:
        obj.terminate()



import asyncio

import pytest

from amrita_core.config import AmritaConfig
from amrita_core.sessions import SessionsManager


class TestSessionsManager:
    """Test SessionsManager class functionality"""

    def setup_method(self):
        """Reset SessionsManager singleton before each test method"""
        # Clear SessionsManager internal state
        SessionsManager._instance = None

    def test_singleton_pattern(self):
        """Test singleton pattern"""
        manager1 = SessionsManager()
        manager2 = SessionsManager()

        assert manager1 is manager2
        assert SessionsManager._instance is manager1

    def test_new_session_creation(self):
        """Test new session creation functionality"""
        manager = SessionsManager()

        # Create new session
        session_id = manager.new_session()

        # Verify session is registered
        assert manager.is_session_registered(session_id)
        assert session_id in [
            data.session_id for data in manager.get_registered_sessions().values()
        ]

        # Verify session resources are initialized
        session_data = manager.get_session_data(session_id)
        assert session_data.tools is not None
        assert session_data.config is not None
        assert session_data.presets is not None

    def test_session_config_operations(self):
        """Test session configuration operations"""
        manager = SessionsManager()

        # Create session
        session_id = manager.new_session()

        # Get initial configuration
        session_data = manager.get_session_data(session_id)
        initial_config = session_data.config
        assert isinstance(initial_config, AmritaConfig)

        # Set new configuration with a valid field
        new_config = AmritaConfig()
        new_config.llm.max_retries = 5  # Using a valid field from LLMConfig
        session_data.config = new_config

        # Verify configuration is updated
        updated_config = manager.get_session_data(session_id).config
        assert updated_config.llm.max_retries == 5

    def test_safe_getters(self):
        """Test safe getters"""
        manager = SessionsManager()

        # Try to get non-existent session
        non_existent_session = "non-existent-session-id"
        try:
            manager.get_session_data(non_existent_session)
            pytest.fail("Expected KeyError was not raised")
        except KeyError:
            pass  # Expected behavior

        # Create session and verify safe getters
        session_id = manager.new_session()
        session_data = manager.get_session_data(session_id)
        assert session_data.config is not None
        assert session_data.tools is not None
        assert session_data.memory is not None

    def test_session_drop(self):
        """Test session drop functionality"""
        manager = SessionsManager()

        # Create session
        session_id = manager.new_session()
        assert manager.is_session_registered(session_id)

        # Drop session
        manager.drop_session(session_id)

        # Verify session is removed
        assert not manager.is_session_registered(session_id)
        assert session_id not in [
            data.session_id for data in manager.get_registered_sessions().values()
        ]

    def test_multiple_sessions(self):
        """Test multiple sessions"""
        manager = SessionsManager()

        # Create multiple sessions
        session_ids = [manager.new_session() for _ in range(3)]

        # Verify all sessions are registered
        for sid in session_ids:
            assert manager.is_session_registered(sid)

        # Verify session count
        assert len(manager.get_registered_sessions()) == 3

        # Set different configurations for each session
        for i, sid in enumerate(session_ids):
            session_data = manager.get_session_data(sid)
            new_config = AmritaConfig()
            new_config.llm.max_retries = 3 + i  # Using a valid field from LLMConfig
            session_data.config = new_config

        # Verify each session has different configuration
        for i, sid in enumerate(session_ids):
            session_data = manager.get_session_data(sid)
            config = session_data.config
            assert config.llm.max_retries == 3 + i

    def test_session_presets(self):
        """Test session presets functionality"""
        manager = SessionsManager()

        # Create session
        session_id = manager.new_session()

        # Verify presets manager exists
        session_data = manager.get_session_data(session_id)
        presets = session_data.presets
        assert presets is not None  # Presets manager is created by default

    def test_session_memory(self):
        """Test session memory functionality"""
        manager = SessionsManager()

        # Create session
        session_id = manager.new_session()

        # Verify memory model exists
        session_data = manager.get_session_data(session_id)
        memory = session_data.memory
        assert memory is not None
        assert hasattr(memory, "messages")
        assert hasattr(memory, "abstract")


@pytest.mark.asyncio
async def test_async_session_operations():
    """Async session operations test"""
    # Simulate async operation
    await asyncio.sleep(0.01)

    manager = SessionsManager()
    session_id = manager.new_session()

    # Simulate operations during async wait
    await asyncio.sleep(0.01)

    # Verify session still exists
    assert manager.is_session_registered(session_id)



from amrita_core.chatmanager import ChatObject
from amrita_core.config import AmritaConfig
from amrita_core.sessions import SessionsManager
from amrita_core.types import (
    MemoryModel,
    ModelPreset,
)


def test_chatobject_uses_session_memory_config_and_default_preset():
    """
    Building a ChatObject with context=None and no config/preset should cause it
    to use the session's memory, config, and default preset.
    """
    session_id = "session-defaults-test"

    # Prepare session-backed memory/config/preset
    session_memory = MemoryModel()
    session_config = AmritaConfig()
    default_preset = ModelPreset(
        model="gpt-3.5-turbo",
        name="session-default",
        api_key="fake-key",
    )

    # this follows the same conventions as the rest of the test suite.
    sessions_manager = SessionsManager()
    sessions_manager.init_session(session_id)
    data = sessions_manager.get_session_data(session_id)
    data.config = session_config
    data.memory = session_memory
    data.presets.set_default_preset(default_preset)

    # No explicit context/config/preset passed in – ChatObject should derive
    # these from the session.
    train = {"role": "system", "content": "system message"}
    user_input = "hello from user"

    chat_obj = ChatObject(
        train=train,
        user_input=user_input,
        context=None,
        session_id=session_id,
        config=None,
        preset=None,
    )

    # ChatObject should be using the session's memory/config/preset
    assert (
        chat_obj.data is session_memory
        or getattr(chat_obj, "memory", None) is session_memory
    )
    assert getattr(chat_obj, "config", None) is session_config
    assert getattr(chat_obj, "preset", None) == default_preset


def test_chatobject_auto_create_session_for_unknown_session_id():
    """
    When auto_create_session=True and the session_id is not registered, the
    ChatObject should trigger session creation and wire itself to that session.
    """
    session_id = "auto-create-session-test"

    # Start with a fresh SessionsManager that does not know about this session_id.
    sessions_manager = SessionsManager()

    train = {"role": "system", "content": "system message"}
    user_input = "hello auto create"
    default_preset = ModelPreset(
        model="gpt-3.5-turbo",
        name="session-default",
        api_key="fake-key",
    )

    # No context/config/preset, unknown session_id, but auto_create_session=True.
    chat_obj = ChatObject(
        train=train,
        user_input=user_input,
        context=None,
        session_id=session_id,
        config=None,
        preset=default_preset,
        auto_create_session=True,
    )

    # The session should have been created
    session = sessions_manager.get_session_data(session_id)
    assert session is not None

    # ChatObject should be wired to the session's memory (and config/preset if present)
    assert isinstance(chat_obj.data, MemoryModel) or isinstance(
        getattr(chat_obj, "memory", None), MemoryModel
    )

    # If the session exposes config/preset defaults, verify ChatObject picked them up.
    session_config = getattr(session, "config", None)
    if session_config is not None:
        assert getattr(chat_obj, "config", None) is session_config



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

from amrita_core.types import USER_INPUT, SendMessageWrap

if TYPE_CHECKING:
    from amrita_core.chatmanager import ChatObject


class EventTypeEnum(str, Enum):
    """
    EventTypeEnum class is used to define and manage different event types.
    It encapsulates the string identifiers of event types, providing a structured way
    to handle and retrieve event types.

    """

    COMPLETION = "COMPLETION"
    Nil = "Nil"
    BEFORE_COMPLETION = "BEFORE_COMPLETION"

    def validate(self, name: str) -> bool:
        return name in self


@dataclass
class Event:
    user_input: USER_INPUT
    original_context: SendMessageWrap
    chat_object: "ChatObject"

    def __post_init__(self):
        # Initialize event type as none
        self._event_type = EventTypeEnum.Nil
        # Validate and store messages using SendMessageWrap
        self._context_messages = self.original_context

    @property
    def event_type(self) -> EventTypeEnum:
        return self._event_type

    @property
    def message(self) -> SendMessageWrap:
        return self._context_messages

    def get_context_messages(self) -> SendMessageWrap:
        return self._context_messages

    def get_event_type(self) -> str:
        raise NotImplementedError

    def get_user_input(self) -> USER_INPUT:
        return self.user_input


@dataclass
class CompletionEvent(Event):
    model_response: str

    def __post_init__(self):
        super().__post_init__()
        # Initialize event type as completion event
        self._event_type = EventTypeEnum.COMPLETION

    @property
    def event_type(self):
        return EventTypeEnum.COMPLETION

    def get_event_type(self) -> str:
        return EventTypeEnum.COMPLETION

    def get_model_response(self) -> str:
        return self.model_response


@dataclass
class PreCompletionEvent(Event):
    def __post_init__(self):
        super().__post_init__()
        self._event_type = EventTypeEnum.BEFORE_COMPLETION

    @property
    def event_type(self) -> EventTypeEnum:
        return self._event_type

    def get_event_type(self) -> str:
        return self._event_type



import random
import time
import typing

from typing_extensions import Self

from .logging import debug_log
from .protocol import AdapterManager
from .tokenizer import hybrid_token_count
from .types import BaseModel, Message, ModelPreset, TextContent, UniResponse

TEST_MSG_PROMPT: Message[list[TextContent]] = Message(
    role="system",
    content=[TextContent(text="You are a helpful assistant.", type="text")],
)
TEST_MSG_USER: Message[list[TextContent]] = Message(
    role="user",
    content=[
        TextContent(text="Hello, please briefly introduce yourself.", type="text")
    ],
)

TEST_MSG_LIST: list[Message[list[TextContent]]] = [
    TEST_MSG_PROMPT,
    TEST_MSG_USER,
]


class PresetReport(BaseModel):
    preset_name: str  # Name of the preset
    preset_data: ModelPreset  # Preset data
    test_input: tuple[Message, Message]  # Test input
    test_output: Message | None  # Test output
    token_prompt: int  # Token count of the prompt
    token_completion: int  # Token count of the completion
    status: bool  # Test result
    message: str  # Test result message
    time_used: float


class MultiPresetManager:
    """
    MultiPresetManager is a class that manages presets.
    """

    _default_preset: ModelPreset | None = None
    _presets: dict[str, ModelPreset]
    _inited = False

    def __init__(self) -> None:
        if not self._inited:
            self._presets = {}
            self._inited = True

    def set_default_preset(self, preset: ModelPreset | str) -> None:
        """
        Set the default preset.
        """
        if isinstance(preset, str):
            preset = self.get_preset(preset)
        if preset.name not in self._presets:
            self.add_preset(preset)
        self._default_preset = preset

    def get_default_preset(self) -> ModelPreset:
        """
        Get the default preset.
        """
        if self._default_preset is None:
            self._default_preset = random.choice(list(self._presets.values()))
        return self._default_preset

    def get_preset(self, name: str) -> ModelPreset:
        """
        Get a preset by name.
        """
        if name not in self._presets:
            raise ValueError(f"Preset {name} not found")
        return self._presets[name]

    def add_preset(self, preset: ModelPreset) -> None:
        """
        Add a preset.
        """
        if preset.name in self._presets:
            raise ValueError(f"Preset {preset.name} already exists")
        self._presets[preset.name] = preset

    def get_all_presets(self) -> list[ModelPreset]:
        """
        Get all presets.
        """
        return list(self._presets.values())

    async def test_single_preset(self, preset: ModelPreset | str) -> PresetReport:
        """Test a single preset for parallel execution"""
        if isinstance(preset, str):
            preset = self.get_preset(preset)
        debug_log(f"Testing preset: {preset.name}...")
        prompt_tokens = hybrid_token_count(
            "".join(
                [typing.cast(TextContent, msg.content[0]).text for msg in TEST_MSG_LIST]
            )
        )

        adapter = AdapterManager().safe_get_adapter(preset.protocol)
        if adapter is None:
            return PresetReport(
                preset_name=preset.name,
                preset_data=preset,
                test_input=(TEST_MSG_PROMPT, TEST_MSG_USER),
                test_output=None,
                token_prompt=prompt_tokens,
                token_completion=0,
                status=False,
                message=f"Undefined protocol adapter: {preset.protocol}",
                time_used=0,
            )

        try:
            time_start = time.time()
            debug_log(f"Calling preset: {preset.name}...")
            data = [  # noqa: RUF015
                i
                async for i in adapter(preset).call_api(TEST_MSG_LIST)
                if isinstance(i, UniResponse)
            ][0]
            time_end = time.time()
            time_delta = time_end - time_start
            debug_log(
                f"Successfully called preset {preset.name}, took {time_delta:.2f} seconds"
            )
            return PresetReport(
                preset_name=preset.name,
                preset_data=preset,
                test_input=(TEST_MSG_PROMPT, TEST_MSG_USER),
                test_output=Message[list[TextContent]](
                    role="assistant",
                    content=[TextContent(type="text", text=data.content)],
                ),
                token_prompt=prompt_tokens,
                token_completion=hybrid_token_count(data.content),
                status=True,
                message="",
                time_used=time_delta,
            )
        except Exception as e:
            debug_log(f"Error occurred while testing preset {preset.name}: {e}")
            return PresetReport(
                preset_name=preset.name,
                preset_data=preset,
                test_input=(TEST_MSG_PROMPT, TEST_MSG_USER),
                test_output=None,
                token_prompt=prompt_tokens,
                token_completion=0,
                status=False,
                message=str(e),
                time_used=0,
            )

    async def test_presets(self) -> typing.AsyncGenerator[PresetReport, None]:
        presets: list[ModelPreset] = self.get_all_presets()
        debug_log(f"Starting to test all presets ({len(presets)} total)...")
        for preset in presets:
            yield await self.test_single_preset(preset)


class PresetManager(MultiPresetManager):
    """
    PresetManager is a singleton class that manages presets.
    """

    _instance = None
    _initialized = False

    def __new__(cls) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        if not self.__class__._initialized:
            super().__init__()
            self.__class__._initialized = True



from __future__ import annotations

import contextlib
import uuid
from dataclasses import dataclass
from typing import TypeVar, overload

from typing_extensions import Self

from amrita_core.config import AmritaConfig
from amrita_core.preset import MultiPresetManager
from amrita_core.tools.mcp import MultiClientManager as ClientManager
from amrita_core.types import MemoryModel

from .tools.manager import MultiToolsManager

T = TypeVar("T")


@dataclass
class SessionData:
    """Container for all session-specific data.

    Attributes:
        session_id: Unique identifier for the session
        memory: Memory model storing conversation history
        tools: Manager for tools available in the session
        presets: Manager for model presets in the session
        mcp: Client manager for MCP (Model Context Protocol) connections
        config: Configuration settings for the session
    """

    session_id: str
    memory: MemoryModel
    tools: MultiToolsManager
    presets: MultiPresetManager
    mcp: ClientManager
    config: AmritaConfig


class SessionsManager:
    """Session manager for managing tools, configurations, and presets for different sessions.

    This manager uses singleton pattern to ensure only one instance exists throughout the application,
    tracking all registered sessions and their associated tools, configurations, and preset managers.
    """

    _session2DataMap: dict[str, SessionData]
    _instance = None
    __marker = object()

    def __new__(cls) -> Self:
        """Implement singleton pattern, create new instance if none exists, otherwise return existing"""
        if cls._instance is None:
            cls._session2DataMap = {}
            cls._instance = super().__new__(cls)
        return cls._instance

    def is_session_registered(self, session_id: str) -> bool:
        """Check if a session is registered.

        Args:
            session_id: The unique identifier for the session

        Returns:
            True if the session is registered, False otherwise
        """
        return session_id in self._session2DataMap

    def get_registered_sessions(self) -> dict[str, SessionData]:
        """
        Get all registered sessions.

        Returns:
            dict[str, SessionData]: A copy of the registered sessions mapping to prevent external modification
        """
        return self._session2DataMap.copy()

    def init_session(self, session_id: str):
        """Initialize resources for a given session.

        Args:
            session_id (str): The unique identifier for the session
        """
        if session_id not in self._session2DataMap:
            self._session2DataMap[session_id] = SessionData(
                session_id,
                MemoryModel(),
                MultiToolsManager(),
                MultiPresetManager(),
                ClientManager(),
                config=AmritaConfig(),
            )

    def set_session(self, data: SessionData):
        """
        Set the session data for a given session ID

        Args:
            data (SessionData): The session data to set
        """
        self._session2DataMap[data.session_id] = data

    @overload
    def get_session_data(self, session_id: str, default: T) -> SessionData | T: ...
    @overload
    def get_session_data(self, session_id: str) -> SessionData: ...
    def get_session_data(
        self, session_id: str, default: T = __marker
    ) -> SessionData | T:
        """Get the session data for a given session ID.

        Args:
            session_id: The unique identifier for the session
            default: Default value to return if session is not found (optional)

        Returns:
            SessionData if session exists, or default value if provided

        Raises:
            KeyError: If session is not found and no default is provided
        """
        if session_id in self._session2DataMap:
            return self._session2DataMap[session_id]
        elif default is not self.__marker:
            return default
        else:
            raise KeyError(f"Session {session_id} not found")

    def new_session(self) -> str:
        """Create a new session with a unique ID and initialize its resources.

        Returns:
            str: The unique identifier for the newly created session
        """
        session_id = uuid.uuid4().hex
        self.init_session(session_id)
        return session_id

    def drop_session(self, session_id: str) -> None:
        """Remove a session and clean up its associated resources.

        Args:
            session_id: The unique identifier for the session to remove
        """
        self._session2DataMap.pop(session_id, None)
        from .chatmanager import chat_manager

        for obj in chat_manager.running_chat_object[session_id]:
            with contextlib.suppress(Exception):
                obj.terminate()
            chat_manager.running_chat_object_id2map.pop(obj.stream_id, None)



from .chatmanager import ChatManager, ChatObject, ChatObjectMeta
from .config import get_config, set_config
from .hook.event import CompletionEvent, PreCompletionEvent
from .hook.matcher import MatcherManager
from .hook.on import on_completion, on_event, on_precompletion
from .libchat import (
    call_completion,
    get_last_response,
    get_tokens,
    text_generator,
    tools_caller,
)
from .logging import debug_log, logger
from .preset import PresetManager, PresetReport
from .sessions import SessionsManager
from .tools import mcp
from .tools.manager import ToolsManager, on_tools, simple_tool
from .tools.models import (
    FunctionDefinitionSchema,
    FunctionParametersSchema,
    FunctionPropertySchema,
    ToolContext,
    ToolData,
    ToolFunctionSchema,
)
from .types import (
    BaseModel,
    Function,
    MemoryModel,
    ModelConfig,
    ModelPreset,
    TextContent,
    ToolCall,
    ToolResult,
    UniResponse,
    UniResponseUsage,
)

__all__ = [
    "BaseModel",
    "ChatManager",
    "ChatObject",
    "ChatObjectMeta",
    "CompletionEvent",
    "Function",
    "FunctionDefinitionSchema",
    "FunctionParametersSchema",
    "FunctionPropertySchema",
    "MatcherManager",
    "MemoryModel",
    "ModelConfig",
    "ModelPreset",
    "PreCompletionEvent",
    "PresetManager",
    "PresetReport",
    "SessionsManager",
    "TextContent",
    "ToolCall",
    "ToolContext",
    "ToolData",
    "ToolFunctionSchema",
    "ToolResult",
    "ToolsManager",
    "UniResponse",
    "UniResponseUsage",
    "call_completion",
    "debug_log",
    "get_config",
    "get_last_response",
    "get_tokens",
    "logger",
    "mcp",
    "on_completion",
    "on_event",
    "on_precompletion",
    "on_tools",
    "set_config",
    "simple_tool",
    "text_generator",
    "tools_caller",
]
__inited: bool = False


def init():
    global __all__, __inited
    if not __inited:
        logger.info("AmritaCore is initalizing......")
        import jieba

        from . import builtins

        __all__ += [builtins.__name__]

        jieba.initialize()


def load_session(session_id: str):
    logger.info("Loading session %s......", session_id)
    sm = SessionsManager()
    sm.init_session(session_id)


async def load_amrita():
    logger.info("Loading AmritaCore......")
    config = get_config()
    if config.function_config.agent_mcp_client_enable:
        logger.info("Loading MCP clients......")
        clients = list(config.function_config.agent_mcp_server_scripts)
        await mcp.ClientManager().initialize_scripts_all(clients)



# TODO:分实例的配置
from __future__ import annotations

import random
import string
from typing import Literal

from pydantic import Field
from typing_extensions import LiteralString

from amrita_core.types import BaseModel


def random_alnum_string(length: int) -> str:
    if length < 0:
        raise ValueError("Length can't be smaller than zero!")

    chars: LiteralString = string.ascii_letters + string.digits

    return "".join(random.choice(chars) for _ in range(length))


class CookieConfig(BaseModel):
    """Amrita Core's cookie config"""

    enable_cookie: bool = Field(
        default=True, description="Whether to enable Cookie leak detection mechanism"
    )
    cookie: str = Field(
        default_factory=lambda: random_alnum_string(16),
        description="Cookie string for security detection",
    )


class FunctionConfig(BaseModel):
    use_minimal_context: bool = Field(
        default=True,
        description="Whether to use minimal context, i.e. system prompt + user's last message (disabling this option will use all context from the message list, which may consume a large amount of Tokens during Agent workflow execution; enabling this option may effectively reduce token usage)",
    )

    tool_calling_mode: Literal["agent", "rag", "none"] = Field(
        default="agent",
        description="Tool calling mode, i.e. whether to use Agent or RAG to call tools",
    )
    agent_tool_call_limit: int = Field(
        default=10, description="Tool call limit in agent mode"
    )
    agent_tool_call_notice: Literal["hide", "notify"] = Field(
        default="hide",
        description="Method of showing tool call status in agent mode, hide to conceal, notify to inform",
    )
    agent_thought_mode: Literal[
        "reasoning", "chat", "reasoning-required", "reasoning-optional"
    ] = Field(
        default="chat",
        description="Thinking mode in agent mode, reasoning mode will first perform reasoning process, then execute tasks; "
        "reasoning-required requires task analysis for each Tool Calling; "
        "reasoning-optional does not require reasoning but allows it; "
        "chat mode executes tasks directly",
    )
    agent_reasoning_hide: bool = Field(
        default=False, description="Whether to hide the thought process in agent mode"
    )
    agent_middle_message: bool = Field(
        default=True,
        description="Whether to allow Agent to send intermediate messages to users in agent mode",
    )
    agent_mcp_client_enable: bool = Field(
        default=False, description="Whether to enable MCP client"
    )
    agent_mcp_server_scripts: list[str] = Field(
        default=[], description="List of MCP server scripts"
    )


class LLMConfig(BaseModel):
    require_tools: bool = Field(
        default=False,
        description="Whether to force at least one tool to be used per call",
    )
    memory_lenth_limit: int = Field(
        default=50, description="Maximum number of messages in memory context"
    )
    max_tokens: int = Field(
        default=100,
        description="Maximum number of tokens generated in a single response",
    )
    tokens_count_mode: Literal["word", "bpe", "char"] = Field(
        default="bpe",
        description="Token counting mode: bpe(subwords)/word(words)/char(characters)",
    )
    enable_tokens_limit: bool = Field(
        default=True, description="Whether to enable context length limits"
    )
    session_tokens_windows: int = Field(
        default=5000, description="Session tokens window size"
    )
    llm_timeout: int = Field(
        default=60, description="API request timeout duration (seconds)"
    )
    auto_retry: bool = Field(
        default=True, description="Automatically retry on request failure"
    )
    max_retries: int = Field(default=3, description="Maximum number of retries")
    enable_memory_abstract: bool = Field(
        default=True,
        description="Whether to enable context memory summarization (will delete context and insert a summary into system instruction)",
    )
    memory_abstract_proportion: float = Field(
        default=15e-2, description="Context summarization proportion (0.15=15%)"
    )
    enable_multi_modal: bool = Field(
        default=True,
        description="Whether to enable multi-modal support (currently only supports image)",
    )


class AmritaConfig(BaseModel):
    function_config: FunctionConfig = Field(
        default_factory=FunctionConfig,
        description="Function configuration",
    )
    llm: LLMConfig = Field(
        default_factory=LLMConfig,
        description="LLM configuration",
    )
    cookie: CookieConfig = Field(
        default_factory=CookieConfig, description="Cookie configuration"
    )


__config = AmritaConfig()
__inited: bool = False


def get_config() -> AmritaConfig:
    """Get the global amrita config

    Raises:
        RuntimeError: Raise it if amrita core is not initialized.

    Returns:
        AmritaConfig: Amrita core config
    """
    if not __inited:
        raise RuntimeError(
            "Global AmritaConfig is not initialized. Please use `set_config` set config first."
        )
    return __config


def set_config(config: AmritaConfig):
    """Override the global config.

    Args:
        config (AmritaConfig): Configuration object to set
    """
    global __config, __inited
    if not __inited:
        __inited = True
    __config = config



import inspect
import re
import typing
from asyncio import iscoroutinefunction
from collections.abc import Awaitable, Callable
from functools import wraps
from typing import Any, get_args, get_origin, get_type_hints, overload

from typing_extensions import Self

from .models import (
    JSON_OBJECT_TYPE,
    FunctionDefinitionSchema,
    FunctionParametersSchema,
    FunctionPropertySchema,
    ToolContext,
    ToolData,
    ToolFunctionSchema,
)

T = typing.TypeVar("T")


class MultiToolsManager:
    _models: dict[str, ToolData]
    _disabled_tools: set[
        str
    ]  # Disabled tools, has_tool and get_tool will not return disabled tools
    _inited = False

    def __init__(self):
        if not self._inited:
            self._models = {}
            self._disabled_tools = set()
            self._inited = True

    def has_tool(self, name: str) -> bool:
        return False if name in self._disabled_tools else name in self._models

    @overload
    def get_tool(self, name: str) -> ToolData | None: ...
    @overload
    def get_tool(self, name: str, default: T) -> ToolData | T: ...
    def get_tool(self, name: str, default: T = None) -> ToolData | T | None:
        if not self.has_tool(name):
            return default
        tool: ToolData = self._models[name]
        return tool if tool.enable_if() else default

    @overload
    def get_tool_meta(self, name: str) -> ToolFunctionSchema | None: ...
    @overload
    def get_tool_meta(self, name: str, default: T) -> ToolFunctionSchema | T: ...
    def get_tool_meta(
        self, name: str, default: T | None = None
    ) -> ToolFunctionSchema | None | T:
        func_data = self.get_tool(name)
        if func_data is None:
            return default
        if isinstance(func_data, ToolData):
            return func_data.data
        return default

    @overload
    def get_tool_func(
        self, name: str, default: T
    ) -> (
        Callable[[dict[str, Any]], Awaitable[str]]
        | Callable[[ToolContext], Awaitable[str | None]]
        | T
    ): ...
    @overload
    def get_tool_func(
        self,
        name: str,
    ) -> (
        Callable[[dict[str, Any]], Awaitable[str]]
        | Callable[[ToolContext], Awaitable[str | None]]
        | None
    ): ...
    def get_tool_func(
        self, name: str, default: T | None = None
    ) -> (
        Callable[[dict[str, Any]], Awaitable[str]]
        | Callable[[ToolContext], Awaitable[str | None]]
        | None
        | T
    ):
        func_data = self.get_tool(name)
        if func_data is None:
            return default
        if isinstance(func_data, ToolData):
            return func_data.func
        return default

    def get_tools(self) -> dict[str, ToolData]:
        return {
            name: data
            for name, data in self._models.items()
            if (name not in self._disabled_tools and data.enable_if())
        }

    def tools_meta(self) -> dict[str, ToolFunctionSchema]:
        return {
            k: v.data
            for k, v in self._models.items()
            if (k not in self._disabled_tools and v.enable_if())
        }

    def tools_meta_dict(self, **kwargs) -> dict[str, dict[str, Any]]:
        return {
            k: v.data.model_dump(**kwargs)
            for k, v in self._models.items()
            if (k not in self._disabled_tools and v.enable_if())
        }

    def register_tool(self, tool: ToolData) -> None:
        if tool.data.function.name not in self._models:
            self._models[tool.data.function.name] = tool
        else:
            raise ValueError(f"Tool {tool.data.function.name} already exists")

    def remove_tool(self, name: str) -> None:
        self._models.pop(name, None)
        if name in self._disabled_tools:
            self._disabled_tools.remove(name)

    def enable_tool(self, name: str) -> None:
        if name in self._disabled_tools:
            self._disabled_tools.remove(name)
        else:
            raise ValueError(f"Tool {name} is not disabled")

    def disable_tool(self, name: str) -> None:
        if self.has_tool(name):
            self._disabled_tools.add(name)
        else:
            raise ValueError(f"Tool {name} does not exist or has been disabled")

    def get_disabled_tools(self) -> list[str]:
        return list(self._disabled_tools)


class ToolsManager(MultiToolsManager):
    _instance = None
    _initialized = False

    def __new__(cls) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        if not self.__class__._initialized:
            super().__init__()
            self.__class__._initialized = True


def _parse_google_docstring(docstring: str | None) -> tuple[str, dict[str, str]]:
    """
    Parse Google-style docstring to extract function description and parameter descriptions.Yes, just like this function's doc.

    Args:
        docstring: The docstring to parse

    Returns:
        A tuple containing (function_description, parameter_descriptions_dict)
    """
    if not docstring:
        return "(no description provided for this tool)", {}

    # Clean up the docstring
    lines = [line.strip() for line in docstring.split("\n") if line.strip()]

    # Find the index where Args section starts
    args_start_idx = -1
    for i, line in enumerate(lines):
        if line.lower().startswith("args:") or line.lower().startswith("参数:"):
            args_start_idx = i
            break

    # Extract function description (everything before Args section)
    if args_start_idx != -1:
        func_desc_lines: list[str] = lines[:args_start_idx]
        func_desc: str = " ".join(func_desc_lines).strip()

        # Extract Args section
        args_lines: list[str] = lines[args_start_idx + 1 :]
    else:
        # No Args section found
        func_desc = " ".join(lines).strip()
        args_lines = []

    # Process Args section
    param_descriptions = {}

    # Pattern to match parameter descriptions in the format:
    # param_name (type): description
    # or
    # param_name: description
    param_pattern = r"^([a-zA-Z_][a-zA-Z0-9_]*)\s*(?:\(([^)]+)\))?\s*:\s*(.*)"

    for line in args_lines:
        # Look for parameter pattern at the beginning of the line or after whitespace
        match = re.match(param_pattern, line)
        if match:
            param_name = match.group(1)
            # param_type = match.group(2)  # We don't need the type since it's in the annotation
            param_desc = match.group(3).strip()

            if param_desc:
                param_descriptions[param_name] = param_desc
            else:
                param_descriptions[param_name] = f"Parameter {param_name}"

    if not func_desc:
        func_desc = "(no description provided for this tool)"

    return func_desc, param_descriptions


def simple_tool(func: Callable[..., Any | Awaitable[Any]]):
    """
    A decorator that creates a ToolData object based on the function signature and annotations.
    It automatically generates parameter descriptions and metadata.

    Here is an example of how to use this decorator:

    ```python
    @simple_tool
    def add(a: int, b: int) -> int:
        \"""Add two numbers together.

        Args:
            a (int): The first number.
            b (int): The second number.

        Returns:
            int: The sum of the two numbers.
        \"""
        return a + b
    ```
    """
    signature: inspect.Signature = inspect.signature(func)

    # Parse Google-style docstring to get function and parameter descriptions
    func_desc, param_descriptions = _parse_google_docstring(func.__doc__)

    # Get the type hints for the function
    type_hints: dict[str, Any] = get_type_hints(
        func, globalns=globals(), localns=locals()
    )

    # Prepare parameters schema
    properties = {}
    required = []

    for param_name, param in signature.parameters.items():
        # Skip 'self' parameter for methods
        if param_name == "self":
            continue

        # Determine parameter type from type hints
        param_type = type_hints.get(param_name)

        # If parameter has no type hint, default to string
        json_type: JSON_OBJECT_TYPE = "string"
        if param_type:
            # Map Python types to JSON schema types
            if hasattr(param_type, "__origin__"):
                origin = get_origin(param_type)
                if origin is not None:
                    if issubclass(origin, list):
                        json_type = "array"
                    elif issubclass(origin, dict):
                        json_type = "object"
                    elif origin is typing.Union:
                        args = get_args(param_type)
                        if type(None) in args:
                            # Handle Optional types - don't add to required
                            pass
                        else:
                            # For Union types, use the first non-None type if available
                            non_none_types = [
                                arg for arg in args if arg is not type(None)
                            ]
                            if non_none_types:
                                param_type = non_none_types[0]
                                json_type = _python_type_to_json_type(param_type)
                else:
                    json_type = _python_type_to_json_type(param_type)
            else:
                json_type = _python_type_to_json_type(param_type)

        # Get parameter description from parsed docstring if available
        param_desc = param_descriptions.get(param_name, f"Parameter {param_name}")

        # Check if parameter is required (no default value)
        is_required = param.default == inspect.Parameter.empty
        if is_required:
            required.append(param_name)
        property_schema = FunctionPropertySchema(type=json_type, description=param_desc)

        properties[param_name] = property_schema

    parameters_schema = FunctionParametersSchema(
        type="object", properties=properties if properties else None, required=required
    )

    function_def = FunctionDefinitionSchema(
        name=func.__name__, description=func_desc, parameters=parameters_schema
    )

    # Create a wrapper function that accepts a dictionary of parameters
    @on_tools(function_def, strict=True)
    @wraps(func)
    async def tool_wrapper(params: dict[str, Any]) -> str:
        bound_args: inspect.BoundArguments = signature.bind(**params)
        bound_args.apply_defaults()

        result = (
            await func(**bound_args.arguments)
            if iscoroutinefunction(func)
            else func(**bound_args.arguments)
        )

        # Convert result to string as expected by the schema
        return str(result)

    return tool_wrapper


def _python_type_to_json_type(python_type: type[Any]) -> JSON_OBJECT_TYPE:
    """Convert Python type to JSON schema type."""
    if python_type is str:
        return "string"
    elif python_type is int:
        return "integer"
    elif python_type is float:
        return "number"
    elif python_type is bool:
        return "boolean"
    elif python_type is list:
        return "array"
    elif python_type is dict:
        return "object"
    else:
        # Default to string for unrecognized types
        return "string"


def on_tools(
    data: FunctionDefinitionSchema,
    custom_run: bool = False,
    strict: bool = False,
    enable_if: Callable[[], bool] = lambda: True,
) -> Callable[
    ...,
    Callable[[dict[str, Any]], Awaitable[str]]
    | Callable[[ToolContext], Awaitable[str | None]],
]:
    """Tool registration decorator

    Args:
        data (FunctionDefinitionSchema): Function metadata
        custom_run (bool, optional): Whether to enable custom run mode. Defaults to False.
        strict (bool, optional): Whether to enable strict mode. Defaults to False.
        show_call (bool, optional): Whether to show tool call. Defaults to True.
    """

    def decorator(
        func: Callable[[dict[str, Any]], Awaitable[str]]
        | Callable[[ToolContext], Awaitable[str | None]],
    ):
        tool_data = ToolData(
            func=func,
            data=ToolFunctionSchema(function=data, type="function", strict=strict),
            custom_run=custom_run,
            enable_if=enable_if,
        )
        ToolsManager().register_tool(tool_data)
        return func

    return decorator



# ref: https://github.com/NoneBot/NoneBot2/blob/main/nonebot/log.py
import inspect
import logging
import os
import sys
from typing import TYPE_CHECKING, Protocol

import loguru

if TYPE_CHECKING:
    from loguru import Logger, Record

logger: "Logger" = loguru.logger

debug: bool = False


class ToStringAble(Protocol):
    def __str__(self) -> str: ...


def debug_log(message: ToStringAble) -> None:
    global debug
    if debug:
        logger.debug(message)


class LoguruHandler(logging.Handler):
    def emit(self, record: logging.LogRecord):
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info, colors=True).log(
            level, record.getMessage()
        )


def default_filter(record: "Record"):
    """Default filter for logging, change level from Environment"""
    log_level = os.environ.get("LOG_LEVEL", "INFO")
    levelno = logger.level(log_level).no if isinstance(log_level, str) else log_level
    return record["level"].no >= levelno


default_format: str = (
    "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "
    "<level>{level: <7}</level> | "
    "<magenta>{name}:{function}:{line}</magenta> | "
    "<level>{message}</level>"
)
"""Default log format"""

logger.remove()
logger_id = logger.add(
    sys.stdout,
    level=0,
    diagnose=False,
    filter=default_filter,
    format=default_format,
)
"""Default log handler id"""


__autodoc__ = {"logger_id": False}



from __future__ import annotations

from collections.abc import Awaitable, Callable
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Generic, Literal, TypeVar

from pydantic import BaseModel, Field, model_validator
from typing_extensions import Self

from ..hook.event import Event
from ..hook.matcher import Matcher

T = TypeVar("T", str, int, float, bool, list, dict)  # JSON type
JOT_T = TypeVar(
    "JOT_T",
    Literal["string"],
    Literal["number"],
    Literal["integer"],
    Literal["boolean"],
    Literal["array"],
    Literal["object"],
)
NUM_T = TypeVar("NUM_T", int, float)
JSON_OBJECT_TYPE = Literal[
    "string",
    "number",
    "integer",
    "boolean",
    "array",
    "object",
]


def on_none(value: Any | None) -> bool:
    """Used for Pydantic's exclude_if

    Args:
        value (Any | None): Value to check

    Returns:
        bool: Returns True when Value is None
    """
    return value is None


def cast_mcp_properties_to_openai(
    property: dict[str, MCP_OBJECT_TYPE],
) -> dict[str, FunctionPropertySchema]:
    """
    Convert MCPPropertySchemaObject dictionary to FunctionPropertySchema objects
    """
    properties_dict: dict[str, FunctionPropertySchema] = {}

    for key, prop in deepcopy(property).items():
        # Convert MCP property type to corresponding FunctionPropertySchema
        converted_prop = _convert_single_property(prop)
        properties_dict[key] = converted_prop

    return properties_dict


def _convert_single_property(mcp_prop: MCP_OBJECT_TYPE) -> FunctionPropertySchema:
    """
    Convert a single MCP_PROPERTY to FunctionPropertySchema
    """
    # Get basic attributes
    description = getattr(
        mcp_prop,
        "description",
        "No description",
    )
    prop_type: JSON_OBJECT_TYPE = mcp_prop.type

    # Prepare base parameters
    base_params = {
        "type": prop_type,
        "description": description,
    }

    # If there are enum values, add to parameters
    if hasattr(mcp_prop, "enum") and mcp_prop.enum is not None:
        base_params["enum"] = mcp_prop.enum

    if isinstance(mcp_prop, MCPPropertySchemaObject):
        # Object type requires recursive conversion of its properties
        if hasattr(mcp_prop, "properties") and mcp_prop.properties:
            obj_properties = {}
            for key, sub_prop in mcp_prop.properties.items():
                obj_properties[key] = _convert_single_property(sub_prop)
            base_params["properties"] = obj_properties
        if hasattr(mcp_prop, "required") and mcp_prop.required:
            base_params["required"] = mcp_prop.required

    elif isinstance(mcp_prop, MCPPropertySchemaArray):
        if hasattr(mcp_prop, "items"):
            base_params["items"] = _convert_single_property(mcp_prop.items)
        if hasattr(mcp_prop, "minItems") and mcp_prop.minItems > 0:
            base_params["minItems"] = mcp_prop.minItems
        if hasattr(mcp_prop, "maxItems") and mcp_prop.maxItems < 100:
            base_params["maxItems"] = mcp_prop.maxItems
        if hasattr(mcp_prop, "uniqueItems"):
            base_params["uniqueItems"] = mcp_prop.uniqueItems

    # For numeric and boolean types, no additional fields are needed since FunctionPropertySchema doesn't include minimum/maximum fields
    return FunctionPropertySchema(**base_params)


class MCPPropertySchema(BaseModel, Generic[JOT_T]):
    """Base structure definition for MCP properties"""

    type: JOT_T = Field(..., description="Parameter type")
    title: str = Field("NO_TITLE", description="Parameter title")
    description: str = Field(
        default="No description", description="Parameter description"
    )
    default: None = Field(
        default=None, description="Parameter default value", exclude_if=on_none
    )
    enum: list[str | int | float] | None = Field(
        default=None, description="Enumerated parameters", exclude_if=on_none
    )


class MCPPropertySchemaString(MCPPropertySchema[Literal["string"]]):
    """MCP property structure for validating string types"""

    pattern: str | None = Field(
        default=None, description="Regular expression", exclude_if=on_none
    )
    minLength: int = Field(default=0, description="Minimum length")
    maxLength: int = Field(default=100, description="Maximum length")


class MCPPropertySchemaNumeric(MCPPropertySchema[JOT_T], Generic[JOT_T, NUM_T]):
    """MCP property structure for validating numeric types"""

    minimum: NUM_T | None = Field(
        default=None, description="Minimum value", exclude_if=on_none
    )
    maximum: NUM_T | None = Field(
        default=None, description="Maximum value", exclude_if=on_none
    )
    exclusiveMinimum: bool | None = Field(
        default=None,
        description="Whether minimum value is exclusive (i.e. 'left open interval')",
        exclude_if=on_none,
    )
    exclusiveMaximum: bool | None = Field(
        default=None,
        description="Whether maximum value is exclusive (i.e. 'right open interval')",
        exclude_if=on_none,
    )


class MCPPropertySchemaInteger(MCPPropertySchemaNumeric[Literal["integer"], int]):
    """MCP property structure for validating integer types"""


class MCPPropertySchemaNumber(MCPPropertySchemaNumeric[Literal["number"], float]):
    """MCP property structure for validating floating point types"""


class MCPPropertySchemaObject(MCPPropertySchema[Literal["object"]]):
    """MCP property structure for validating object types"""

    properties: dict[str, MCP_OBJECT_TYPE] = Field(
        ..., description="Parameter property definitions", exclude_if=on_none
    )
    required: list[str] = Field(
        default_factory=list, description="List of required parameters"
    )

    @model_validator(mode="after")
    def validator(self) -> Self:
        if self.required:
            for req in self.required:
                if req not in self.properties:
                    raise ValueError(
                        f"Required property '{req}' not found in properties."
                    )
        return self


class MCPPropertySchemaArray(MCPPropertySchema[Literal["array"]]):
    """MCP property structure for validating array types"""

    items: MCP_OBJECT_TYPE = Field(
        ..., description="Parameter property definition", exclude_if=on_none
    )
    maxItems: int = Field(default=100, description="Maximum number of elements")
    minItems: int = Field(default=0, description="Minimum number of elements")
    uniqueItems: bool = Field(
        default=False,
        description="Whether array elements must be unique, when type is array, this parameter defaults to False",
    )


class MCPPropertySchemaBoolean(MCPPropertySchema[Literal["boolean"]]):
    """MCP property structure for validating boolean types"""


MCP_OBJECT_TYPE = (
    MCPPropertySchemaObject
    | MCPPropertySchemaString
    | MCPPropertySchemaNumber
    | MCPPropertySchemaInteger
    | MCPPropertySchemaArray
    | MCPPropertySchemaBoolean
)


class MCPToolSchema(BaseModel):
    """Define the structure of MCP tools"""

    name: str = Field(..., description="Tool name")
    description: str = Field("No description", description="Tool description")
    inputSchema: MCPPropertySchemaObject = Field(
        ..., description="Tool parameter definition"
    )


class FunctionPropertySchema(BaseModel, Generic[T]):
    """Property for validating function arguments"""

    type: Literal[JSON_OBJECT_TYPE] | list[JSON_OBJECT_TYPE] = Field(
        ..., description="Parameter type"
    )
    description: str = Field("No description", description="Parameter description")
    enum: list[T] | None = Field(
        default=None, description="Enumerated parameters", exclude_if=on_none
    )
    properties: dict[str, FunctionPropertySchema] | None = Field(
        default=None,
        description="Parameter property definitions, only valid when parameter type is object",
        exclude_if=on_none,
    )
    items: FunctionPropertySchema | None = Field(
        default=None,
        description="Used only when type='array', defines array element type",
        exclude_if=on_none,
    )
    minItems: int | None = Field(
        default=None,
        description="Used only when type='array', defines minimum array length",
        exclude_if=on_none,
    )
    maxItems: int | None = Field(
        default=None,
        description="Used only when type='array', defines maximum number of array elements",
        exclude_if=on_none,
    )
    uniqueItems: bool | None = Field(
        default=None,
        description="Whether array elements must be unique, when type is array, this parameter defaults to False",
        exclude_if=on_none,
    )
    required: list[str] | None = Field(
        default=None,
        description="Parameter property definitions, only valid when parameter type is object",
        exclude_if=on_none,
    )

    @model_validator(mode="after")
    def validator(self) -> Self:
        if self.type == "object":
            if self.properties is None:
                raise ValueError("When type is object, properties must be set.")
            elif self.required is None:
                self.required = []
            if any(
                i is not None
                for i in (self.maxItems, self.minItems, self.uniqueItems, self.items)
            ):
                raise ValueError(
                    "When type is object, `maxItems`,`minItems`,`uniqueItems`,`Items` must be None."
                )
        elif self.type == "array":
            if self.items is None:
                raise ValueError("When type is array, items must be set.")
            elif self.minItems is not None and self.minItems < 0:
                raise ValueError("minItems must be greater than or equal to 0.")
            elif self.maxItems is not None and self.maxItems < 0:
                raise ValueError("maxItems must be greater than or equal to 0.")
            elif (
                self.maxItems is not None
                and self.minItems is not None
                and self.maxItems < self.minItems
            ):
                raise ValueError("maxItems must be greater than or equal to minItems.")
            elif self.uniqueItems is None:
                self.uniqueItems = False

        return self


class FunctionParametersSchema(BaseModel):
    """Validate function parameter structure"""

    type: Literal["object"] = Field(..., description="Parameter type")
    properties: dict[str, FunctionPropertySchema] | None = Field(
        default=None, description="Parameter property definitions", exclude_if=on_none
    )

    required: list[str] = Field(
        default_factory=list, description="List of required parameters"
    )


class FunctionDefinitionSchema(BaseModel):
    """Validate function definition structure"""

    name: str = Field(..., description="Function name")
    description: str = Field(..., description="Function description")
    parameters: FunctionParametersSchema = Field(
        ..., description="Function parameter definition"
    )


class ToolFunctionSchema(BaseModel):
    """Validate complete function field structure"""

    function: FunctionDefinitionSchema = Field(..., description="Function definition")
    type: Literal["function"] = "function"
    strict: bool = Field(default=False, description="Whether in strict mode")


ToolChoice = Literal["none", "auto", "required"] | ToolFunctionSchema


@dataclass
class ToolContext:
    data: dict[str, Any] = field()
    event: Event = field()
    matcher: Matcher = field()


class ToolData(BaseModel):
    """Data model for registering Tools"""

    data: ToolFunctionSchema = Field(..., description="Tool metadata")
    func: (
        Callable[[dict[str, Any]], Awaitable[str]]
        | Callable[[ToolContext], Awaitable[str | None]]
    ) = Field(..., description="Tool function")
    custom_run: bool = Field(
        default=False,
        description="Whether to customize execution; if enabled, passes Context class instead of dict and does not enforce return value.",
    )
    enable_if: Callable[[], bool] = Field(
        default=lambda: True,
        description="Whether to enable this tool",
    )



import json
import random
from asyncio import Lock
from collections.abc import Awaitable, Callable, Iterable
from contextlib import nullcontext
from copy import deepcopy
from typing import Any, overload

from fastmcp import Client
from fastmcp.client.client import CallToolResult
from fastmcp.client.transports import ClientTransportT
from mcp.types import TextContent
from typing_extensions import Self
from zipp import Path

from amrita_core.logging import logger

from .manager import MultiToolsManager, ToolsManager
from .models import (
    FunctionDefinitionSchema,
    FunctionParametersSchema,
    MCPToolSchema,
    ToolData,
    ToolFunctionSchema,
    cast_mcp_properties_to_openai,
)

MCP_SERVER_SCRIPT_TYPE = ClientTransportT


class NOT_GIVEN:
    pass


class MCPClient:
    """Reusable MCP Client"""

    mcp_client: Client | None = None
    server_script: MCP_SERVER_SCRIPT_TYPE

    def __init__(
        self,
        server_script: MCP_SERVER_SCRIPT_TYPE,
        # headers: dict | None = None,
    ):
        self.mcp_client = None
        self.server_script: MCP_SERVER_SCRIPT_TYPE = server_script
        self.tools: list[MCPToolSchema] = []
        self.openai_tools: list[ToolFunctionSchema] = []

    async def __aenter__(self) -> Self:
        await self._connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._close()

    async def simple_call(self, tool_name: str, data: dict[str, Any]) -> str:
        """Call MCP tool
        Args:
            tool_name (str): tool name
            data (dict[str, Any]): tool parameters
        """

        try:
            if self.mcp_client is None:
                await self._connect()
            assert self.mcp_client is not None
            response: CallToolResult = await self.mcp_client.call_tool(tool_name, data)
            ct: list[TextContent] = [
                i for i in response.content if isinstance(i, TextContent)
            ]
            return "".join([f"{i.text}\n\n" for i in ct])
        except Exception as e:
            logger.opt(exception=e, colors=True).error(
                f"Failed to call tool:{tool_name}, because {e}."
            )
            return json.dumps({"success": False, "error": str(e)})
        finally:
            await self._close()

    async def _connect(self, update_tools: bool = False):
        """Connect to MCP Server
        Args:
            update_tools (bool, optional): whether to update the tool list. Defaults to False.
        """
        if self.mcp_client is not None:
            raise RuntimeError("MCP Server is already connected!")

        server_script = self.server_script
        self.mcp_client = Client(server_script)
        await self.mcp_client.__aenter__()
        logger.info(f"Successfully connected to MCP Server@{server_script}")
        if not self.tools or update_tools:
            self.tools = [
                MCPToolSchema.model_validate(i.model_dump())
                for i in await self.mcp_client.list_tools()
            ]
            logger.info(f"Available tools: {[tool.name for tool in self.tools]}")
            self._cast_tool_to_openai()

    def _format_tools_for_openai(self):
        """Convert MCP tool format to OpenAI tool format"""
        openai_tools: list[ToolFunctionSchema] = [
            ToolFunctionSchema(
                strict=True,
                type="function",
                function=FunctionDefinitionSchema(
                    name=tool.name,
                    description=tool.description or f"Running tool named: {tool.name}",
                    parameters=FunctionParametersSchema(
                        type="object",
                        required=tool.inputSchema.required,
                        properties=cast_mcp_properties_to_openai(
                            property=tool.inputSchema.properties
                        ),
                    ),
                ),
            )
            for tool in self.tools
        ]
        return openai_tools

    def _cast_tool_to_openai(self) -> None:
        self.openai_tools = self._format_tools_for_openai()

    def get_tools(self) -> list[ToolFunctionSchema]:
        """Get MCP tool list"""
        return self.openai_tools

    def get_original_tools(self) -> list[MCPToolSchema]:
        """Get original MCP tool list"""
        return self.tools

    async def _close(self):
        """Close connection"""
        if self.mcp_client:
            await self.mcp_client.__aexit__(None, None, None)
            self.mcp_client = None


class MultiClientManager:
    clients: list[MCPClient]
    script_to_clients: dict[str, MCPClient]
    name_to_clients: dict[str, MCPClient]  # Map from FunctionName to MCPClient
    tools_remapping: dict[
        str, str
    ]  # Remap duplicate tools for SuggarChat (original_name->remapped_name)
    reversed_remappings: dict[
        str, str
    ]  # Reverse mapping (remapped_name->original_name)
    tools_manager: MultiToolsManager = ToolsManager()
    _lock: Lock
    _is_initialized = False  # Whether ToolsMapping is ready
    _initted = False

    def __init__(self, tools_manager: MultiToolsManager | None = None) -> None:
        if not self._initted:
            self._initted = True
            self.tools_manager = tools_manager or self.tools_manager
            self.clients = []
            self.name_to_clients = {}
            self.tools_remapping = {}
            self.reversed_remappings = {}
            self.script_to_clients = {}
            self._lock = Lock()

    def get_client_by_script(self, server_script: MCP_SERVER_SCRIPT_TYPE) -> MCPClient:
        """Get MCP Client (without operating stored MCP Server)
        Args:
            server_script (str, optional): MCP Server script path (or URI).
        """
        return MCPClient(server_script)

    async def get_client_by_tool_name(self, tool_name: str) -> MCPClient:
        """Get MCP Client by tool name
        Args:
            tool_name (str): tool name
        """
        async with self._lock:
            name = self.tools_remapping.get(tool_name) or tool_name
            if name in self.name_to_clients:
                return self.name_to_clients[name]
            raise RuntimeError(
                f"Tool not found: {tool_name}{f' (remapped from `{name}`)' if name != tool_name else ''}"
            )

    def _tools_wrapper(
        self, tool_name: str
    ) -> Callable[[dict[str, Any]], Awaitable[str]]:
        async def tools_runner(data: dict[str, Any]) -> str:
            client: MCPClient = await self.get_client_by_tool_name(tool_name)
            return await client.simple_call(tool_name, data)

        return tools_runner

    @overload
    def register_only(self, *, client: MCPClient) -> Self:
        """Register MCP Server only, without initialization"""
        ...

    @overload
    def register_only(self, *, server_script: MCP_SERVER_SCRIPT_TYPE) -> Self:
        """Register MCP Server only, without initialization"""
        ...

    def register_only(
        self,
        *,
        server_script: MCP_SERVER_SCRIPT_TYPE | None = None,
        client: MCPClient | None = None,
    ) -> Self:
        """Register MCP Server only, without initialization"""
        if client is not None:
            self.clients.append(client)
        elif server_script is not None:
            client = MCPClient(server_script)
            self.clients.append(client)
        else:
            raise ValueError("Please provide MCP Server script or MCP Client")
        return self

    async def update_tools(self, client: MCPClient):
        tools = client.get_tools()
        async with self._lock:
            for tool in tools:
                name = tool.function.name
                self.tools_manager.remove_tool(name)
                self.name_to_clients.pop(name, None)
                if remap := self.tools_remapping.pop(name, None):
                    self.reversed_remappings.pop(remap, None)
        await self._load_this(client)

    async def initialize_scripts_all(
        self, scripts: Iterable[MCP_SERVER_SCRIPT_TYPE]
    ) -> Self:
        for script in scripts:
            await self.initialize_this(script)
        return self

    async def initialize_this(
        self, server_script: MCP_SERVER_SCRIPT_TYPE, fail_then_raise: bool = False
    ) -> Self:
        """Register and initialize single MCP Server"""
        client: MCPClient = self.get_client_by_script(server_script)
        async with self._lock:
            try:
                await self._load_this(client)
            except Exception as e:
                logger.error(f"Failed to initialize MCP Server@{server_script}: {e}")
                if fail_then_raise:
                    raise
            else:
                self.clients.append(client)
        return self

    async def _load_this(self, client: MCPClient, fail_then_raise=True):
        try:
            tools_remapping_tmp = {}
            reversed_remappings_tmp = {}
            name_to_clients_tmp = {}
            tm = self.tools_manager
            async with client as c:
                tools = deepcopy(c.get_tools())
                for tool in tools:
                    if (
                        tool.function.name in self.tools_remapping
                        or tool.function.name in self.name_to_clients
                    ):
                        logger.warning(
                            f"{client}@{client.server_script} has a tool named {tool.function.name}, which is already registered, the old tool will be replaced."
                        )
                    name_to_clients_tmp[tool.function.name] = client
                    origin_name = tool.function.name
                    if tm.has_tool(tool.function.name):
                        remapped_name = (
                            f"referred_{random.randint(1, 100)}_{tool.function.name}"
                        )
                        logger.warning(
                            f"Tool already exists: {tool.function.name}, it will be remapped to: {remapped_name}"
                        )
                        tools_remapping_tmp[origin_name] = remapped_name
                        reversed_remappings_tmp[remapped_name] = origin_name
                        tool.function.name = remapped_name

                    tm.register_tool(
                        ToolData(
                            data=tool,
                            func=self._tools_wrapper(origin_name),
                        )
                    )

        except Exception as e:
            if fail_then_raise:
                raise
            logger.opt(exception=e, colors=True).error(
                f"Failed to connect to MCP Server@{client.server_script}: {e}"
            )
        else:
            logger.info(f"Successfully loaded MCP Server@{client.server_script}")
            self.tools_remapping.update(tools_remapping_tmp)
            self.reversed_remappings.update(reversed_remappings_tmp)
            self.name_to_clients.update(name_to_clients_tmp)
            if isinstance(client.server_script, str | Path):
                server_script = str(client.server_script)
                self.script_to_clients[server_script] = client

    async def reinitalize_all(self):
        async with self._lock:
            for client in deepcopy(self.clients):
                await self.unregister_client(client.server_script, False)
                self.register_only(client=client)
                await self._load_this(client, fail_then_raise=False)

    async def initialize_all(self, lock: bool = True):
        """Connect to all MCP Servers"""
        async with self._lock if lock else nullcontext():
            for client in self.clients:
                await self._load_this(client, False)
            self._is_initialized = True

    async def unregister_client(self, script_name: str | Path, lock: bool = True):
        """Unregister an MCP Server"""
        tools_manager = self.tools_manager
        async with self._lock if lock else nullcontext():
            script_name = str(script_name)
            if script_name in self.script_to_clients:
                client = self.script_to_clients.pop(script_name)
                for tool in client.openai_tools:
                    name = tool.function.name
                    tools_manager.remove_tool(name)
                    self.name_to_clients.pop(name, None)
                    if remap := self.tools_remapping.pop(name, None):
                        tools_manager.remove_tool(remap)
                        self.reversed_remappings.pop(remap, None)
                for idx, client in enumerate(self.clients):
                    if client.server_script == script_name:
                        self.clients.pop(idx)
                        break


class ClientManager(MultiClientManager):
    _instance = None
    _initialized = False

    def __new__(cls) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        if not self.__class__._initialized:
            super().__init__()
            self.__class__._initialized = True



import inspect
from collections.abc import Awaitable, Callable
from copy import deepcopy
from types import FrameType
from typing import (
    Any,
    ClassVar,
    TypeAlias,
)

from pydantic import BaseModel, Field
from typing_extensions import Self

from amrita_core.config import AmritaConfig
from amrita_core.logging import debug_log, logger

from .event import Event
from .exception import BlockException, CancelException, PassException

ChatException: TypeAlias = BlockException | CancelException | PassException


class FunctionData(BaseModel, arbitrary_types_allowed=True):
    function: Callable[..., Awaitable[Any]] = Field(...)
    signature: inspect.Signature = Field(...)
    frame: FrameType = Field(...)
    priority: int = Field(...)
    block: bool = Field(...)
    matcher: Any = Field(...)


class EventRegistry:
    _instance = None
    __event_handlers: ClassVar[dict[str, list[FunctionData]]] = {}

    def __new__(cls) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def register_handler(self, event_type: str, data: FunctionData):
        self.__event_handlers.setdefault(event_type, []).append(data)

    def get_handlers(self, event_type: str) -> list[FunctionData]:
        self.__event_handlers.setdefault(event_type, [])
        self.__event_handlers[event_type].sort(key=lambda x: x.priority, reverse=False)
        return self.__event_handlers[event_type]

    def _all(self) -> dict[str, list[FunctionData]]:
        return self.__event_handlers


class Matcher:
    def __init__(self, event_type: str, priority: int = 10, block: bool = True):
        """Constructor, initialize Matcher object.
        Args:
            event_type (str): Event type
            priority (int, optional): Priority. Defaults to 10.
            block (bool, optional): Whether to block subsequent events. Defaults to True.
        """
        if priority <= 0:
            raise ValueError("Event priority cannot be zero or negative!")

        self.event_type = event_type
        self.priority = priority
        self.block = block

    def append_handler(self, func: Callable[..., Awaitable[Any]]):
        frame = inspect.currentframe()
        assert frame is not None, "Frame is None!!!"
        func_data = FunctionData(
            function=func,
            signature=inspect.signature(func),
            frame=frame,
            priority=self.priority,
            block=self.block,
            matcher=self,
        )
        EventRegistry().register_handler(self.event_type, func_data)

    def handle(self):
        """
        Event handler registration function
        """

        def wrapper(
            func: Callable[..., Awaitable[Any]],
        ):
            self.append_handler(func)
            return func

        return wrapper

    def stop_process(self):
        """
        Stop the event flow within the current chat plugin and immediately stop the current handler.
        """
        raise BlockException()

    def cancel_matcher(self):
        """
        Stop event processing within the current chat plugin and cancel.
        """
        raise CancelException()

    def pass_event(self):
        """
        Ignore the current handler and continue processing the next one.
        """
        raise PassException()


class MatcherManager:
    """
    Event handling manager.
    """

    @staticmethod
    async def trigger_event(*args, **kwargs) -> None:
        """
        Trigger a specific type of event and call all registered event handlers for that type.

        Parameters:
        - event: Event object containing event-related data.
        - **kwargs: Keyword arguments passed to the dependency injection system.
        - *args: Variable arguments passed to the dependency injection system.
        """
        event: Event | None = None
        config: AmritaConfig | None = None
        for i in args:
            if isinstance(i, Event):
                event = i
            elif isinstance(i, AmritaConfig):
                config = i
        if not event:
            raise RuntimeError("No event found in args")
        elif not config:
            raise RuntimeError("No config found in args")
        event_type = event.get_event_type()  # Get event type
        priority_tmp = 0
        debug_log(f"Running matchers for event: {event_type}!")
        # Check if there are handlers for this event type
        if matcher_list := EventRegistry().get_handlers(event_type):
            for matcher in matcher_list:
                if matcher.priority != priority_tmp:
                    priority_tmp = matcher.priority
                    debug_log(f"Running matchers for priority {priority_tmp}......")

                signature = matcher.signature
                frame = matcher.frame
                line_number = frame.f_lineno
                file_name = frame.f_code.co_filename
                handler = matcher.function
                session_args = [matcher.matcher, *args]
                session_kwargs = {**deepcopy(kwargs)}

                args_types = {k: v.annotation for k, v in signature.parameters.items()}
                filtered_args_types = {
                    k: v for k, v in args_types.items() if v is not inspect._empty
                }
                if args_types != filtered_args_types:
                    failed_args = list(args_types.keys() - filtered_args_types.keys())
                    logger.warning(
                        f"Matcher {matcher.function.__name__} (File: {file_name}: Line {frame.f_lineno!s}) has untyped parameters!"
                        + f"(Args:{''.join(i + ',' for i in failed_args)}).Skipping......"
                    )
                    continue
                new_args = []
                used_indices = set()
                for param_type in filtered_args_types.values():
                    for i, arg in enumerate(session_args):
                        if i in used_indices:
                            continue
                        if isinstance(arg, param_type):
                            new_args.append(arg)
                            used_indices.add(i)
                            break

                # Get keyword argument type annotations
                kwparams = signature.parameters
                f_kwargs = {  # TODO: kwparams dependent support
                    param_name: session_kwargs[param.annotation]
                    for param_name, param in kwparams.items()
                    if param.annotation in session_kwargs
                }
                if len(new_args) != len(list(filtered_args_types)):
                    continue

                # Call the handler

                try:
                    logger.info(f"Starting to run Matcher: '{handler.__name__}'")

                    await handler(*new_args, **f_kwargs)
                except PassException:
                    logger.info(
                        f"Matcher '{handler.__name__}'(~{file_name}:{line_number}) was skipped"
                    )
                    continue
                except CancelException:
                    logger.info("Cancelled Matcher processing")
                    return
                except BlockException:
                    break
                except Exception as e:
                    logger.opt(exception=e, colors=True).error(
                        f"An error occurred while running '{handler.__name__}'({file_name}:{line_number}) "
                    )

                    continue
                finally:
                    logger.info(f"Handler {handler.__name__} finished")
                    if matcher.block:
                        break
        else:
            logger.warning(
                f"No registered Matcher for {event_type} event, skipping processing."
            )



from amrita_core.tools.models import (
    FunctionDefinitionSchema,
    FunctionParametersSchema,
    FunctionPropertySchema,
    ToolFunctionSchema,
)

PROCESS_MESSAGE_TOOL = FunctionDefinitionSchema(
    name="processing_message",
    description="Describe what the agent is currently doing and express the agent's internal thoughts to the user. Use this when you need to communicate your current actions or internal reasoning to the user, not for general completion.",
    parameters=FunctionParametersSchema(
        type="object",
        properties={
            "content": FunctionPropertySchema(
                description="Message content, describe in the tone of system instructions what you are doing or interacting with the user.",
                type="string",
            ),
        },
        required=["content"],
    ),
)
PROCESS_MESSAGE = ToolFunctionSchema(
    type="function",
    function=PROCESS_MESSAGE_TOOL,
    strict=True,
)
STOP_TOOL = ToolFunctionSchema(
    type="function",
    function=FunctionDefinitionSchema(
        name="agent_stop",
        description="Call this tool to indicate that you have gathered enough information and are ready to formulate the final answer to the user.\n"
        + " After calling this, you should NOT call any other tools, but directly provide the completion",
        parameters=FunctionParametersSchema(
            type="object",
            properties={
                "result": FunctionPropertySchema(
                    type="string",
                    description="Simply illustrate what you did during the chat task.(Optional)",
                )
            },
            required=[],
        ),
    ),
    strict=True,
)

REASONING_TOOL = ToolFunctionSchema(
    type="function",
    function=FunctionDefinitionSchema(
        name="think_and_reason",
        description="Think about what you should do next, always call this tool to think when completing a tool call.",
        parameters=FunctionParametersSchema(
            type="object",
            properties={
                "content": FunctionPropertySchema(
                    description="What you should do next",
                    type="string",
                ),
            },
            required=["content"],
        ),
    ),
    strict=True,
)
